from functools import cached_property
from typing import Any, Dict, Optional, Tuple, Union

from vyper import ast as vy_ast
from vyper.abi_types import ABIType
from vyper.ast.identifiers import validate_identifier
from vyper.exceptions import (
    CompilerPanic,
    InvalidLiteral,
    InvalidOperation,
    NamespaceCollision,
    StructureException,
    UnknownAttribute,
)
from vyper.semantics.analysis.levenshtein_utils import get_levenshtein_error_suggestions
from vyper.semantics.data_locations import DataLocation


# Some fake type with an overridden `compare_type` which accepts any RHS
# type of type `type_`
class _GenericTypeAcceptor:
    def __repr__(self):
        return f"GenericTypeAcceptor({self.type_})"

    def __init__(self, type_):
        self.type_ = type_

    def compare_type(self, other):
        if isinstance(other, self.type_):
            return True
        # compare two GenericTypeAcceptors -- they are the same if the base
        # type is the same
        return isinstance(other, self.__class__) and other.type_ == self.type_

    def to_dict(self):
        # this shouldn't really appear in the AST type annotations, but it's
        # there for certain string literals which don't have a known type. this
        # should be fixed soon by improving type inference. for now just put
        # *something* in the AST.
        return {"generic": self.type_.typeclass}


class VyperType:
    """
    Base class for vyper types.

    Attributes
    ----------
    _id : str
        The name of the type.
    _as_array: bool, optional
        If `True`, this type can be used as the base member for an array.
    _valid_literal : Tuple
        A tuple of Vyper ast classes that may be assigned this type.
    _invalid_locations : Tuple
        A tuple of invalid `DataLocation`s for this type
    _is_prim_word: bool, optional
        This is a word type like uint256, int8, bytesM or address
    _supports_external_calls: bool, optional
        Whether or not this type supports external calls. Currently
        limited to `InterfaceT`s
    _attribute_in_annotation: bool, optional
        Whether or not this type can be attributed in a type
        annotation, like IFoo.SomeType. Currently limited to
        `InterfaceT`s.
    """

    typeclass: str = None  # type: ignore

    _id: str  # rename to `_name`
    _type_members: Optional[Dict] = None
    _valid_literal: Tuple = ()
    _invalid_locations: Tuple = ()
    _is_prim_word: bool = False
    _equality_attrs: Optional[Tuple] = None
    _is_array_type: bool = False
    _is_bytestring: bool = False  # is it a bytes or a string?

    _as_array: bool = False  # rename to something like can_be_array_member
    _as_hashmap_key: bool = False

    _supports_external_calls: bool = False
    _attribute_in_annotation: bool = False

    size_in_bytes = 32  # default; override for larger types

    decl_node: Optional[vy_ast.VyperNode] = None

    def __init__(self, members: Optional[Dict] = None) -> None:
        self.members: Dict = {}

        # add members that are on the class instance.
        if self._type_members is not None:
            for k, v in self._type_members.items():
                # for builtin members like `contract.address` -- skip namespace
                # validation, as it introduces a dependency cycle
                self.add_member(k, v)

        members = members or {}
        for k, v in members.items():
            self.add_member(k, v)

    def _get_equality_attrs(self):
        return tuple(getattr(self, attr) for attr in self._equality_attrs)

    def __hash__(self):
        return hash(self._get_equality_attrs())

    def __eq__(self, other):
        if self is other:
            return True
        return (
            type(self) is type(other) and self._get_equality_attrs() == other._get_equality_attrs()
        )

    def __lt__(self, other):
        return self.abi_type.selector_name() < other.abi_type.selector_name()

    # return a dict suitable for serializing in the AST
    def to_dict(self):
        ret = {"name": self._id}
        if self.decl_node is not None:
            ret["type_decl_node"] = self.decl_node.get_id_dict()
        if self.typeclass is not None:
            ret["typeclass"] = self.typeclass

        # use dict ctor to block duplicates
        return dict(**self._addl_dict_fields(), **ret)

    # for most types, this is a reasonable implementation, but it can
    # be overridden as needed.
    def _addl_dict_fields(self):
        keys = self._equality_attrs or ()
        ret = {}
        for k in keys:
            if k.startswith("_"):
                continue
            v = getattr(self, k)
            if hasattr(v, "to_dict"):
                v = v.to_dict()
            ret[k] = v
        return ret

    @cached_property
    def _as_darray(self):
        return self._as_array

    @property
    def getter_signature(self):
        return (), self

    # TODO not sure if this is a great idea.
    @classmethod
    def any(cls):
        return _GenericTypeAcceptor(cls)

    @property
    def abi_type(self) -> ABIType:
        """
        The ABI type corresponding to this type
        """
        raise CompilerPanic("Method must be implemented by the inherited class")

    def get_size_in(self, location: DataLocation):
        if location in (DataLocation.STORAGE, DataLocation.TRANSIENT):
            return self.storage_size_in_words
        if location == DataLocation.MEMORY:
            return self.memory_bytes_required
        if location == DataLocation.CODE:
            return self.memory_bytes_required

        raise CompilerPanic(f"unreachable: invalid location {location}")  # pragma: nocover

    @property
    def memory_bytes_required(self) -> int:
        # alias for API compatibility with codegen
        return self.size_in_bytes

    @property
    def storage_size_in_words(self) -> int:
        # consider renaming if other word-addressable address spaces are
        # added to EVM or exist in other arches
        """
        Returns the number of words required to allocate in storage for
        this type
        """
        r = self.memory_bytes_required
        if r % 32 != 0:
            raise CompilerPanic("Memory bytes must be multiple of 32")
        return r // 32

    @property
    def canonical_abi_type(self) -> str:
        """
        The canonical name of this type. Used for ABI types and generating function signatures.
        """
        return self.abi_type.selector_name()

    def to_abi_arg(self, name: str = "") -> dict[str, Any]:
        """
        The JSON ABI description of this type. Note for complex types,
        the implementation is overridden to be compliant with the spec:
        https://docs.soliditylang.org/en/v0.8.14/abi-spec.html#json
        > An object with members name, type and potentially components
          describes a typed variable. The canonical type is determined
          until a tuple type is reached and the string description up to
          that point is stored in type prefix with the word tuple, i.e.
          it will be tuple followed by a sequence of [] and [k] with
          integers k. The components of the tuple are then stored in the
          member components, which is of array type and has the same
          structure as the top-level object except that indexed is not
          allowed there.
        """
        return {"name": name, "type": self.canonical_abi_type}

    # convenience method for erroring out of invalid ast ops
    def _raise_invalid_op(
        self,
        # TODO maybe make these AST classes inherit from "HasOperator"
        node: Union[vy_ast.UnaryOp, vy_ast.BinOp, vy_ast.AugAssign, vy_ast.Compare, vy_ast.BoolOp],
    ) -> None:
        raise InvalidOperation(f"Cannot perform {node.op.description} on {self}", node.op)

    def validate_comparator(self, node: vy_ast.Compare) -> None:
        """
        Validate a comparator for this type.

        Arguments
        ---------
        node : Compare
            Vyper ast node of the comparator to be validated.

        Returns
        -------
        None. A failed validation must raise an exception.
        """
        if not isinstance(node.op, (vy_ast.Eq, vy_ast.NotEq)):
            self._raise_invalid_op(node)

    def validate_numeric_op(
        self, node: Union[vy_ast.UnaryOp, vy_ast.BinOp, vy_ast.AugAssign]
    ) -> None:
        """
        Validate a numeric operation for this type.

        Arguments
        ---------
        node : UnaryOp | BinOp | AugAssign
            Vyper ast node of the numeric operation to be validated.

        Returns
        -------
        None. A failed validation must raise an exception.
        """
        self._raise_invalid_op(node)

    def validate_boolean_op(self, node: vy_ast.BoolOp) -> None:
        """
        Validate a boolean operation for this type.

        Arguments
        ---------
        node : BoolOp
            Vyper ast node of the boolean operation to be validated.

        Returns
        -------
        None. A failed validation must raise an exception.
        """
        self._raise_invalid_op(node)

    def validate_literal(self, node: vy_ast.Constant) -> None:
        """
        Validate that a literal node can be annotated with this type

        Arguments
        ---------
        node : VyperNode
            `Constant` Vyper ast node, or a list or tuple of constants.
        """
        if not isinstance(node, self._valid_literal) or not isinstance(node, vy_ast.Constant):
            # should not reach here, by paths into validate_literal.
            raise InvalidLiteral(f"Invalid literal for {self._id}", node)

    def validate_index_type(self, node: vy_ast.Subscript) -> None:
        raise StructureException(f"Not an indexable type: '{self}'", node)

    def compare_type(self, other: "VyperType") -> bool:
        """
        Compare this type object against another type object.

        Failed comparisons must return `False`, not raise an exception.

        This method does *not* test for type equality, it is a type
        checker function, it should have the meaning: "an expr of type
        <other> can be assigned to an expr of type <self>."

        Arguments
        ---------
        other: VyperType
            Another type object to be compared against this one.

        Returns
        -------
        bool
            Indicates if the types are equivalent.
        """
        return isinstance(other, type(self))

    def fetch_call_return(self, node: vy_ast.Call) -> Optional["VyperType"]:
        """
        Validate a call to this type and return the result.

        This method must raise if the type is not callable, or the call arguments
        are not valid.

        Arguments
        ---------
        node : Call
            Vyper ast node of call action to validate.

        Returns
        -------
        VyperType, optional
            Type generated as a result of the call.
        """
        raise StructureException(f"{self} is not callable", node)

    @classmethod
    def get_subscripted_type(self, node: vy_ast.VyperNode) -> None:
        """
        Return the type of a subscript expression, e.g. x[1]

        Arguments
        ---------
        node: Index
            Vyper ast node from the `slice` member of a Subscript node

        Returns
        -------
        VyperType
            Type object for value at the given index.
        """
        raise StructureException(f"'{self}' cannot be indexed into", node)

    def _check_add_member(self, name):
        if (prev_type := self.members.get(name)) is not None:
            msg = f"Member '{name}' already exists in {self}"
            raise NamespaceCollision(msg, prev_decl=prev_type.decl_node)

    def add_member(self, name: str, type_: "VyperType") -> None:
        validate_identifier(name)
        self._check_add_member(name)
        self.members[name] = type_

    def get_member(self, key: str, node: vy_ast.VyperNode) -> "VyperType":
        if key in self.members:
            return self.members[key]

        # special error message for types with no members
        if not self.members:
            raise StructureException(f"{self} instance does not have members", node)

        hint = get_levenshtein_error_suggestions(key, self.members, 0.3)
        raise UnknownAttribute(f"{self} has no member '{key}'.", node, hint=hint)

    def __repr__(self):
        return self._id


class KwargSettings:
    # convenience class which holds metadata about how to process kwargs.
    # contains the `default` value for the kwarg as a python value, and a
    # flag `require_literal`, which, when True, indicates that the kwarg
    # must be set to a compile-time constant at any call site.
    # (note that the kwarg processing machinery will return a
    # Python value instead of an AST or IRnode in this case).
    def __init__(self, typ, default, require_literal=False):
        self.typ = typ
        self.default = default
        self.require_literal = require_literal


class _VoidType(VyperType):
    _id = "(void)"


# sentinel for function calls which return nothing
VOID_TYPE = _VoidType()


def map_void(typ: Optional[VyperType]) -> VyperType:
    if typ is None:
        return VOID_TYPE
    return typ


# A type type. Used internally for types which can live in expression
# position, ex. constructors (events, interfaces and structs), and also
# certain builtins which take types as parameters
class TYPE_T(VyperType):
    def __init__(self, typedef):
        super().__init__()

        self.typedef = typedef

    def to_dict(self):
        return {"type_t": self.typedef.to_dict()}

    def __repr__(self):
        return f"type({self.typedef})"

    def check_modifiability_for_call(self, node, modifiability):
        if hasattr(self.typedef, "_ctor_modifiability_for_call"):
            return self.typedef._ctor_modifiability_for_call(node, modifiability)
        raise StructureException("Value is not callable", node)

    # dispatch into ctor if it's called
    def fetch_call_return(self, node):
        if hasattr(self.typedef, "_ctor_call_return"):
            return self.typedef._ctor_call_return(node)
        raise StructureException("Value is not callable", node)

    def infer_arg_types(self, node, expected_return_typ=None):
        if hasattr(self.typedef, "_ctor_arg_types"):
            return self.typedef._ctor_arg_types(node)
        raise StructureException("Value is not callable", node)

    def infer_kwarg_types(self, node):
        if hasattr(self.typedef, "_ctor_kwarg_types"):
            return self.typedef._ctor_kwarg_types(node)
        raise StructureException("Value is not callable", node)

    # dispatch into get_type_member if it's dereferenced, ex.
    # MyFlag.FOO
    def get_member(self, key, node):
        if hasattr(self.typedef, "get_type_member"):
            return self.typedef.get_type_member(key, node)
        raise UnknownAttribute("Value is not attributable", node)


def is_type_t(x: VyperType, t: type) -> bool:
    return isinstance(x, TYPE_T) and isinstance(x.typedef, t)



from vyper.venom.analysis.dfg import DFGAnalysis
from vyper.venom.analysis.liveness import LivenessAnalysis
from vyper.venom.passes.base_pass import IRPass


class RemoveUnusedVariablesPass(IRPass):
    def run_pass(self):
        removeList = set()

        self.analyses_cache.request_analysis(LivenessAnalysis)

        for bb in self.function.get_basic_blocks():
            for i, inst in enumerate(bb.instructions[:-1]):
                if inst.volatile:
                    continue
                next_liveness = bb.instructions[i + 1].liveness
                if (inst.output and inst.output not in next_liveness) or inst.opcode == "nop":
                    removeList.add(inst)

            bb.instructions = [inst for inst in bb.instructions if inst not in removeList]

        self.analyses_cache.invalidate_analysis(DFGAnalysis)



from typing import Optional

from vyper.venom.basicblock import IRInstruction, IRLabel, IROperand
from vyper.venom.function import IRFunction


class IRContext:
    functions: dict[IRLabel, IRFunction]
    ctor_mem_size: Optional[int]
    immutables_len: Optional[int]
    data_segment: list[IRInstruction]
    last_label: int

    def __init__(self) -> None:
        self.functions = {}
        self.ctor_mem_size = None
        self.immutables_len = None
        self.data_segment = []
        self.last_label = 0

    def add_function(self, fn: IRFunction) -> None:
        fn.ctx = self
        self.functions[fn.name] = fn

    def create_function(self, name: str) -> IRFunction:
        label = IRLabel(name, True)
        fn = IRFunction(label, self)
        self.add_function(fn)
        return fn

    def get_function(self, name: IRLabel) -> IRFunction:
        if name in self.functions:
            return self.functions[name]
        raise Exception(f"Function {name} not found in context")

    def get_next_label(self, suffix: str = "") -> IRLabel:
        if suffix != "":
            suffix = f"_{suffix}"
        self.last_label += 1
        return IRLabel(f"{self.last_label}{suffix}")

    def chain_basic_blocks(self) -> None:
        """
        Chain basic blocks together. This is necessary for the IR to be valid, and is done after
        the IR is generated.
        """
        for fn in self.functions.values():
            fn.chain_basic_blocks()

    def append_data(self, opcode: str, args: list[IROperand]) -> None:
        """
        Append data
        """
        self.data_segment.append(IRInstruction(opcode, args))  # type: ignore

    def as_graph(self) -> str:
        s = ["digraph G {"]
        for fn in self.functions.values():
            s.append(fn.as_graph(True))
            s.append("\n")
        s.append("}")
        return "\n".join(s)

    def __repr__(self) -> str:
        s = ["IRContext:"]
        for fn in self.functions.values():
            s.append(fn.__repr__())
            s.append("\n")

        if len(self.data_segment) > 0:
            s.append("\nData segment:")
            for inst in self.data_segment:
                s.append(f"{inst}")

        return "\n".join(s)



import contextlib
import sys
import zipfile
from pathlib import Path

import pytest

from vyper.cli.vyper_compile import compile_files
from vyper.cli.vyper_json import compile_json
from vyper.compiler.input_bundle import FilesystemInputBundle
from vyper.compiler.output_bundle import OutputBundle
from vyper.compiler.phases import CompilerData
from vyper.utils import sha256sum


def test_combined_json_keys(chdir_tmp_path, make_file):
    make_file("bar.vy", "")

    combined_keys = {
        "bytecode",
        "bytecode_runtime",
        "blueprint_bytecode",
        "abi",
        "source_map",
        "source_map_runtime",
        "layout",
        "method_identifiers",
        "userdoc",
        "devdoc",
    }
    compile_data = compile_files(["bar.vy"], ["combined_json"])

    assert set(compile_data.keys()) == {Path("bar.vy"), "version"}
    assert set(compile_data[Path("bar.vy")].keys()) == combined_keys


def test_invalid_root_path():
    with pytest.raises(FileNotFoundError):
        compile_files([], [], paths=["path/that/does/not/exist"])


CONTRACT_CODE = """
{import_stmt}

@external
def foo() -> {alias}.FooStruct:
    return {alias}.FooStruct(foo_=13)

@external
def bar(a: address) -> {alias}.FooStruct:
    return extcall {alias}(a).bar()
"""

INTERFACE_CODE = """
struct FooStruct:
    foo_: uint256

@external
def foo() -> FooStruct:
    ...

@external
def bar() -> FooStruct:
    ...
"""


SAME_FOLDER_IMPORT_STMT = [
    ("import IFoo as IFoo", "IFoo"),
    ("import contracts.IFoo as IFoo", "IFoo"),
    ("from . import IFoo", "IFoo"),
    ("from contracts import IFoo", "IFoo"),
    ("from ..contracts import IFoo", "IFoo"),
    ("from . import IFoo as FooBar", "FooBar"),
    ("from contracts import IFoo as FooBar", "FooBar"),
    ("from ..contracts import IFoo as FooBar", "FooBar"),
]


@pytest.mark.parametrize("import_stmt,alias", SAME_FOLDER_IMPORT_STMT)
def test_import_same_folder(import_stmt, alias, chdir_tmp_path, make_file):
    foo = "contracts/foo.vy"
    make_file("contracts/foo.vy", CONTRACT_CODE.format(import_stmt=import_stmt, alias=alias))
    make_file("contracts/IFoo.vyi", INTERFACE_CODE)

    assert compile_files([foo], ["combined_json"]) is not None


SUBFOLDER_IMPORT_STMT = [
    ("import other.IFoo as IFoo", "IFoo"),
    ("import contracts.other.IFoo as IFoo", "IFoo"),
    ("from other import IFoo", "IFoo"),
    ("from contracts.other import IFoo", "IFoo"),
    ("from .other import IFoo", "IFoo"),
    ("from ..contracts.other import IFoo", "IFoo"),
    ("from other import IFoo as FooBar", "FooBar"),
    ("from contracts.other import IFoo as FooBar", "FooBar"),
    ("from .other import IFoo as FooBar", "FooBar"),
    ("from ..contracts.other import IFoo as FooBar", "FooBar"),
]


@pytest.mark.parametrize("import_stmt, alias", SUBFOLDER_IMPORT_STMT)
def test_import_subfolder(import_stmt, alias, chdir_tmp_path, make_file):
    foo = make_file(
        "contracts/foo.vy", (CONTRACT_CODE.format(import_stmt=import_stmt, alias=alias))
    )
    make_file("contracts/other/IFoo.vyi", INTERFACE_CODE)

    assert compile_files([foo], ["combined_json"]) is not None


OTHER_FOLDER_IMPORT_STMT = [
    ("import interfaces.IFoo as IFoo", "IFoo"),
    ("from interfaces import IFoo", "IFoo"),
    ("from ..interfaces import IFoo", "IFoo"),
    ("from interfaces import IFoo as FooBar", "FooBar"),
    ("from ..interfaces import IFoo as FooBar", "FooBar"),
]


@pytest.mark.parametrize("import_stmt, alias", OTHER_FOLDER_IMPORT_STMT)
def test_import_other_folder(import_stmt, alias, tmp_path, make_file):
    foo = make_file("contracts/foo.vy", CONTRACT_CODE.format(import_stmt=import_stmt, alias=alias))
    make_file("interfaces/IFoo.vyi", INTERFACE_CODE)

    assert compile_files([foo], ["combined_json"], paths=[tmp_path]) is not None


def test_import_parent_folder(tmp_path, make_file):
    foo = make_file(
        "contracts/baz/foo.vy",
        CONTRACT_CODE.format(import_stmt="from ... import IFoo", alias="IFoo"),
    )
    make_file("IFoo.vyi", INTERFACE_CODE)

    assert compile_files([foo], ["combined_json"], paths=[tmp_path]) is not None

    # perform relative import outside of base folder
    compile_files([foo], ["combined_json"], paths=[tmp_path / "contracts"])


def test_import_search_paths(chdir_tmp_path, make_file):
    contract_code = CONTRACT_CODE.format(import_stmt="from utils import IFoo", alias="IFoo")
    contract_filename = "dir1/baz/foo.vy"
    interface_filename = "dir2/utils/IFoo.vyi"
    make_file(interface_filename, INTERFACE_CODE)
    make_file(contract_filename, contract_code)

    assert compile_files([contract_filename], ["combined_json"], paths=["dir2"]) is not None


META_IMPORT_STMT = [
    "import ISelf as ISelf",
    "import contracts.ISelf as ISelf",
    "from . import ISelf",
    "from contracts import ISelf",
]


@pytest.mark.parametrize("import_stmt", META_IMPORT_STMT)
def test_import_self_interface(import_stmt, tmp_path, make_file):
    interface_code = """
struct FooStruct:
    foo_: uint256

@external
def know_thyself(a: address) -> FooStruct:
    ...

@external
def be_known() -> FooStruct:
    ...
    """
    code = f"""
{import_stmt}

@external
def know_thyself(a: address) -> ISelf.FooStruct:
    return extcall ISelf(a).be_known()

@external
def be_known() -> ISelf.FooStruct:
    return ISelf.FooStruct(foo_=42)
    """
    make_file("contracts/ISelf.vyi", interface_code)
    meta = make_file("contracts/Self.vy", code)

    assert compile_files([meta], ["combined_json"], paths=[tmp_path]) is not None


# implement IFoo in another contract for fun
@pytest.mark.parametrize("import_stmt_foo,alias", SAME_FOLDER_IMPORT_STMT)
def test_another_interface_implementation(import_stmt_foo, alias, tmp_path, make_file):
    baz_code = f"""
{import_stmt_foo}

@external
def foo(a: address) -> {alias}.FooStruct:
    return extcall {alias}(a).foo()

@external
def bar(_foo: address) -> {alias}.FooStruct:
    return extcall {alias}(_foo).bar()
    """
    make_file("contracts/IFoo.vyi", INTERFACE_CODE)
    baz = make_file("contracts/Baz.vy", baz_code)

    assert compile_files([baz], ["combined_json"], paths=[tmp_path]) is not None


def test_local_namespace(make_file, chdir_tmp_path):
    # interface code namespaces should be isolated
    # all of these contract should be able to compile together
    codes = [
        "import foo as FooBar",
        "import bar as FooBar",
        "import foo as BarFoo",
        "import bar as BarFoo",
    ]
    struct_def = """
struct FooStruct:
    foo_: uint256

    """

    paths = []
    for i, code in enumerate(codes):
        code += struct_def
        filename = f"code{i}.vy"
        make_file(filename, code)
        paths.append(filename)

    for file_name in ("foo.vyi", "bar.vyi"):
        make_file(file_name, INTERFACE_CODE)

    assert compile_files(paths, ["combined_json"]) is not None


def test_compile_outside_root_path(tmp_path, make_file):
    # absolute paths relative to "."
    make_file("ifoo.vyi", INTERFACE_CODE)
    foo = make_file("foo.vy", CONTRACT_CODE.format(import_stmt="import ifoo as IFoo", alias="IFoo"))

    assert compile_files([foo], ["combined_json"], paths=None) is not None


def test_import_library(tmp_path, make_file):
    library_source = """
@internal
def foo() -> uint256:
    return block.number + 1
    """

    contract_source = """
import lib

@external
def foo() -> uint256:
    return lib.foo()
    """

    make_file("lib.vy", library_source)
    contract_file = make_file("contract.vy", contract_source)

    assert compile_files([contract_file], ["combined_json"], paths=[tmp_path]) is not None


@contextlib.contextmanager
def mock_sys_path(path):
    try:
        sys.path.append(path)
        yield
    finally:
        sys.path.pop()


@pytest.fixture
def input_files(tmp_path_factory, make_file, chdir_tmp_path):
    library_source = """
@internal
def foo() -> uint256:
    return block.number + 1
    """
    json_source = """
[
  {
    "stateMutability": "nonpayable",
    "type": "function",
    "name": "test_json",
    "inputs": [ { "name": "", "type": "uint256" } ],
    "outputs": [ { "name": "", "type": "uint256" } ]
  }
]
    """
    contract_source = """
import lib
import jsonabi

@external
def foo() -> uint256:
    return lib.foo()

@external
def bar(x: uint256) -> uint256:
    return extcall jsonabi(msg.sender).test_json(x)
    """
    tmpdir = tmp_path_factory.mktemp("fake-package")
    with open(tmpdir / "lib.vy", "w") as f:
        f.write(library_source)
    with open(tmpdir / "jsonabi.json", "w") as f:
        f.write(json_source)

    contract_file = make_file("contract.vy", contract_source)

    return (tmpdir, tmpdir / "lib.vy", tmpdir / "jsonabi.json", contract_file)


def test_import_sys_path(input_files):
    tmpdir, _, _, contract_file = input_files
    with mock_sys_path(tmpdir):
        assert compile_files([contract_file], ["combined_json"]) is not None


def test_archive_output(input_files):
    tmpdir, _, _, contract_file = input_files
    search_paths = [".", tmpdir]

    s = compile_files([contract_file], ["archive"], paths=search_paths)
    archive_bytes = s[contract_file]["archive"]

    archive_path = Path("foo.zip")
    with archive_path.open("wb") as f:
        f.write(archive_bytes)

    assert zipfile.is_zipfile(archive_path)

    # compare compiling the two input bundles
    out = compile_files([contract_file], ["integrity", "bytecode"], paths=search_paths)
    out2 = compile_files([archive_path], ["integrity", "bytecode"])
    assert out[contract_file] == out2[archive_path]


def test_archive_b64_output(input_files):
    tmpdir, _, _, contract_file = input_files
    search_paths = [".", tmpdir]

    out = compile_files(
        [contract_file], ["archive_b64", "integrity", "bytecode"], paths=search_paths
    )

    archive_b64 = out[contract_file].pop("archive_b64")

    archive_path = Path("foo.zip.b64")
    with archive_path.open("w") as f:
        f.write(archive_b64)

    # compare compiling the two input bundles
    out2 = compile_files([archive_path], ["integrity", "bytecode"])
    assert out[contract_file] == out2[archive_path]


def test_solc_json_output(input_files):
    tmpdir, _, _, contract_file = input_files
    search_paths = [".", tmpdir]

    out = compile_files([contract_file], ["solc_json"], paths=search_paths)

    json_input = out[contract_file]["solc_json"]

    # check that round-tripping solc_json thru standard json produces
    # the same as compiling directly
    json_out = compile_json(json_input)["contracts"]["contract.vy"]
    json_out_bytecode = json_out["contract"]["evm"]["bytecode"]["object"]

    out2 = compile_files([contract_file], ["integrity", "bytecode"], paths=search_paths)

    assert out2[contract_file]["bytecode"] == json_out_bytecode


# maybe this belongs in tests/unit/compiler?
def test_integrity_sum(input_files):
    tmpdir, library_file, jsonabi_file, contract_file = input_files
    search_paths = [".", tmpdir]

    out = compile_files([contract_file], ["integrity"], paths=search_paths)

    with library_file.open() as f, contract_file.open() as g, jsonabi_file.open() as h:
        library_contents = f.read()
        contract_contents = g.read()
        jsonabi_contents = h.read()

    contract_hash = sha256sum(contract_contents)
    library_hash = sha256sum(library_contents)
    jsonabi_hash = sha256sum(jsonabi_contents)
    expected = sha256sum(contract_hash + sha256sum(library_hash) + jsonabi_hash)
    assert out[contract_file]["integrity"] == expected


# does this belong in tests/unit/compiler?
def test_archive_search_path(tmp_path_factory, make_file, chdir_tmp_path):
    lib1 = """
x: uint256
    """
    lib2 = """
y: uint256
    """
    dir1 = tmp_path_factory.mktemp("dir1")
    dir2 = tmp_path_factory.mktemp("dir2")
    make_file(dir1 / "lib.vy", lib1)
    make_file(dir2 / "lib.vy", lib2)

    main = """
import lib
    """
    pwd = Path(".")
    make_file(pwd / "main.vy", main)
    for search_paths in ([pwd, dir1, dir2], [pwd, dir2, dir1]):
        input_bundle = FilesystemInputBundle(search_paths)
        file_input = input_bundle.load_file("main.vy")

        # construct CompilerData manually
        compiler_data = CompilerData(file_input, input_bundle)
        output_bundle = OutputBundle(compiler_data)

        used_dir = search_paths[-1].stem  # either dir1 or dir2
        assert output_bundle.used_search_paths == [".", "0/" + used_dir]



import base64
import warnings
from collections import deque
from pathlib import PurePath

from vyper.ast import ast_to_dict
from vyper.codegen.ir_node import IRnode
from vyper.compiler.output_bundle import SolcJSONWriter, VyperArchiveWriter
from vyper.compiler.phases import CompilerData
from vyper.compiler.utils import build_gas_estimates
from vyper.evm import opcodes
from vyper.exceptions import VyperException
from vyper.ir import compile_ir
from vyper.semantics.types.function import FunctionVisibility, StateMutability
from vyper.typing import StorageLayout
from vyper.utils import vyper_warn
from vyper.warnings import ContractSizeLimitWarning


def build_ast_dict(compiler_data: CompilerData) -> dict:
    ast_dict = {
        "contract_name": str(compiler_data.contract_path),
        "ast": ast_to_dict(compiler_data.vyper_module),
    }
    return ast_dict


def build_annotated_ast_dict(compiler_data: CompilerData) -> dict:
    annotated_ast_dict = {
        "contract_name": str(compiler_data.contract_path),
        "ast": ast_to_dict(compiler_data.annotated_vyper_module),
    }
    return annotated_ast_dict


def build_devdoc(compiler_data: CompilerData) -> dict:
    return compiler_data.natspec.devdoc


def build_userdoc(compiler_data: CompilerData) -> dict:
    return compiler_data.natspec.userdoc


def build_solc_json(compiler_data: CompilerData) -> str:
    # request bytecode to ensure the input compiles through all the
    # compilation passes, emit warnings if there are any issues
    # (this allows use cases like sending a bug reproduction while
    # still alerting the user in the common case that they didn't
    # mean to have a bug)
    try:
        _ = compiler_data.bytecode
    except VyperException as e:
        vyper_warn(
            f"Exceptions encountered during code generation (but producing output anyway): {e}"
        )
    writer = SolcJSONWriter(compiler_data)
    writer.write()
    return writer.output()


def build_archive(compiler_data: CompilerData) -> bytes:
    # ditto
    try:
        _ = compiler_data.bytecode
    except VyperException as e:
        vyper_warn(
            f"Exceptions encountered during code generation (but producing archive anyway): {e}"
        )
    writer = VyperArchiveWriter(compiler_data)
    writer.write()
    return writer.output()


def build_archive_b64(compiler_data: CompilerData) -> str:
    return base64.b64encode(build_archive(compiler_data)).decode("ascii")


def build_integrity(compiler_data: CompilerData) -> str:
    return compiler_data.compilation_target._metadata["type"].integrity_sum


def build_external_interface_output(compiler_data: CompilerData) -> str:
    interface = compiler_data.annotated_vyper_module._metadata["type"].interface
    stem = PurePath(compiler_data.contract_path).stem
    # capitalize words separated by '_'
    # ex: test_interface.vy -> TestInterface
    name = "".join([x.capitalize() for x in stem.split("_")])
    out = f"\n# External Interfaces\ninterface {name}:\n"

    for func in interface.functions.values():
        if func.visibility == FunctionVisibility.INTERNAL or func.name == "__init__":
            continue
        args = ", ".join([f"{arg.name}: {arg.typ}" for arg in func.arguments])
        return_value = f" -> {func.return_type}" if func.return_type is not None else ""
        mutability = func.mutability.value
        out = f"{out}    def {func.name}({args}){return_value}: {mutability}\n"

    return out


def build_interface_output(compiler_data: CompilerData) -> str:
    interface = compiler_data.annotated_vyper_module._metadata["type"].interface
    out = ""

    if interface.events:
        out = "# Events\n\n"
        for event in interface.events.values():
            encoded_args = "\n    ".join(f"{name}: {typ}" for name, typ in event.arguments.items())
            out = f"{out}event {event.name}:\n    {encoded_args if event.arguments else 'pass'}\n"

    if interface.functions:
        out = f"{out}\n# Functions\n\n"
        for func in interface.functions.values():
            if func.visibility == FunctionVisibility.INTERNAL or func.name == "__init__":
                continue
            if func.mutability != StateMutability.NONPAYABLE:
                out = f"{out}@{func.mutability.value}\n"
            args = ", ".join([f"{arg.name}: {arg.typ}" for arg in func.arguments])
            return_value = f" -> {func.return_type}" if func.return_type is not None else ""
            out = f"{out}@external\ndef {func.name}({args}){return_value}:\n    ...\n\n"

    return out


def build_bb_output(compiler_data: CompilerData) -> IRnode:
    return compiler_data.venom_functions[0]


def build_bb_runtime_output(compiler_data: CompilerData) -> IRnode:
    return compiler_data.venom_functions[1]


def build_cfg_output(compiler_data: CompilerData) -> str:
    return compiler_data.venom_functions[0].as_graph()


def build_cfg_runtime_output(compiler_data: CompilerData) -> str:
    return compiler_data.venom_functions[1].as_graph()


def build_ir_output(compiler_data: CompilerData) -> IRnode:
    if compiler_data.show_gas_estimates:
        IRnode.repr_show_gas = True
    return compiler_data.ir_nodes


def build_ir_runtime_output(compiler_data: CompilerData) -> IRnode:
    if compiler_data.show_gas_estimates:
        IRnode.repr_show_gas = True
    return compiler_data.ir_runtime


def _ir_to_dict(ir_node):
    # Currently only supported with IRnode and not VenomIR
    if not isinstance(ir_node, IRnode):
        return
    args = ir_node.args
    if len(args) > 0 or ir_node.value == "seq":
        return {ir_node.value: [_ir_to_dict(x) for x in args]}
    return ir_node.value


def build_ir_dict_output(compiler_data: CompilerData) -> dict:
    return _ir_to_dict(compiler_data.ir_nodes)


def build_ir_runtime_dict_output(compiler_data: CompilerData) -> dict:
    return _ir_to_dict(compiler_data.ir_runtime)


def build_metadata_output(compiler_data: CompilerData) -> dict:
    sigs = compiler_data.function_signatures

    def _var_rec_dict(variable_record):
        ret = vars(variable_record).copy()
        ret["typ"] = str(ret["typ"])
        if ret["data_offset"] is None:
            del ret["data_offset"]
        for k in ("blockscopes", "defined_at", "encoding"):
            del ret[k]
        ret["location"] = ret["location"].name
        return ret

    def _to_dict(func_t):
        ret = vars(func_t).copy()
        ret["return_type"] = str(ret["return_type"])
        ret["_ir_identifier"] = func_t._ir_info.ir_identifier

        for attr in ("mutability", "visibility"):
            ret[attr] = ret[attr].name.lower()

        # e.g. {"x": vy_ast.Int(..)} -> {"x": 1}
        ret["default_values"] = {
            k: val.node_source_code for k, val in func_t.default_values.items()
        }

        for attr in ("positional_args", "keyword_args"):
            args = ret[attr]
            ret[attr] = {arg.name: str(arg.typ) for arg in args}

        ret["frame_info"] = vars(func_t._ir_info.frame_info).copy()
        del ret["frame_info"]["frame_vars"]  # frame_var.pos might be IR, cannot serialize

        keep_keys = {
            "name",
            "return_type",
            "positional_args",
            "keyword_args",
            "default_values",
            "frame_info",
            "mutability",
            "visibility",
            "_ir_identifier",
            "nonreentrant_key",
        }
        ret = {k: v for k, v in ret.items() if k in keep_keys}
        return ret

    return {"function_info": {name: _to_dict(sig) for (name, sig) in sigs.items()}}


def build_method_identifiers_output(compiler_data: CompilerData) -> dict:
    module_t = compiler_data.annotated_vyper_module._metadata["type"]
    functions = module_t.exposed_functions

    return {k: hex(v) for fn_t in functions for k, v in fn_t.method_ids.items()}


def build_abi_output(compiler_data: CompilerData) -> list:
    module_t = compiler_data.annotated_vyper_module._metadata["type"]
    _ = compiler_data.ir_runtime  # ensure _ir_info is generated

    abi = module_t.interface.to_toplevel_abi_dict()
    if compiler_data.show_gas_estimates:
        # Add gas estimates for each function to ABI
        gas_estimates = build_gas_estimates(compiler_data.function_signatures)
        for func in abi:
            try:
                func_signature = func["name"]
            except KeyError:
                # constructor and fallback functions don't have a name
                continue

            func_name, _, _ = func_signature.partition("(")
            # This check ensures we skip __init__ since it has no estimate
            if func_name in gas_estimates:
                func["gas"] = gas_estimates[func_name]
    return abi


def build_asm_output(compiler_data: CompilerData) -> str:
    return _build_asm(compiler_data.assembly)


def build_layout_output(compiler_data: CompilerData) -> StorageLayout:
    # in the future this might return (non-storage) layout,
    # for now only storage layout is returned.
    return compiler_data.storage_layout


def _build_asm(asm_list):
    output_string = ""
    in_push = 0
    for node in asm_list:
        if isinstance(node, list):
            output_string += "{ " + _build_asm(node) + "} "
            continue

        if in_push > 0:
            assert isinstance(node, int), node
            output_string += hex(node)[2:].rjust(2, "0")
            if in_push == 1:
                output_string += " "
            in_push -= 1
        else:
            output_string += str(node) + " "

            if isinstance(node, str) and node.startswith("PUSH") and node != "PUSH0":
                assert in_push == 0
                in_push = int(node[4:])
                output_string += "0x"

    return output_string


def _build_node_identifier(ast_node):
    assert ast_node.module_node is not None, type(ast_node)
    return (ast_node.module_node.source_id, ast_node.node_id)


def _build_source_map_output(compiler_data, bytecode, pc_maps):
    """
    Generate source map output in various formats. Note that integrations
    are encouraged to use pc_ast_map since the information it provides is
    a superset of the other formats, and the other types are included
    for legacy reasons.
    """
    # sort the pc maps alphabetically
    # CMC 2024-03-09 is this really necessary?
    out = {}
    for k in sorted(pc_maps.keys()):
        out[k] = pc_maps[k]

    ast_map = out.pop("pc_raw_ast_map")

    assert isinstance(ast_map, dict)  # lint
    if 0 not in ast_map:
        # tag it with source id
        ast_map[0] = compiler_data.annotated_vyper_module

    pc_pos_map = {k: compile_ir.getpos(v) for (k, v) in ast_map.items()}
    node_id_map = {k: _build_node_identifier(v) for (k, v) in ast_map.items()}
    compressed_map = _compress_source_map(ast_map, out["pc_jump_map"], len(bytecode))
    out["pc_pos_map_compressed"] = compressed_map
    out["pc_pos_map"] = pc_pos_map
    out["pc_ast_map"] = node_id_map
    # hint to consumers what the fields in pc_ast_map mean
    out["pc_ast_map_item_keys"] = ("source_id", "node_id")
    return out


def build_source_map_output(compiler_data: CompilerData) -> dict:
    bytecode, pc_maps = compile_ir.assembly_to_evm(
        compiler_data.assembly, insert_compiler_metadata=False
    )
    return _build_source_map_output(compiler_data, bytecode, pc_maps)


def build_source_map_runtime_output(compiler_data: CompilerData) -> dict:
    bytecode, pc_maps = compile_ir.assembly_to_evm(
        compiler_data.assembly_runtime, insert_compiler_metadata=False
    )
    return _build_source_map_output(compiler_data, bytecode, pc_maps)


# generate a solidity-style source map. this functionality is deprecated
# in favor of pc_ast_map, and may not be maintained to the same level
# as pc_ast_map.
def _compress_source_map(ast_map, jump_map, bytecode_size):
    ret = []

    jump_map = jump_map.copy()
    ast_map = ast_map.copy()

    for pc in range(bytecode_size):
        if pc in ast_map:
            ast_node = ast_map.pop(pc)
            # ast_node.src conveniently has the current position in
            # the correct, compressed format
            current_pos = [ast_node.src]
        else:
            current_pos = ["-1:-1:-1"]

        if pc in jump_map:
            jump_type = jump_map.pop(pc)
            current_pos.append(jump_type)

        ret.append(":".join(str(i) for i in current_pos))

    assert len(ast_map) == 0, ast_map
    assert len(jump_map) == 0, jump_map

    return ";".join(ret)


def build_bytecode_output(compiler_data: CompilerData) -> str:
    return f"0x{compiler_data.bytecode.hex()}"


def build_blueprint_bytecode_output(compiler_data: CompilerData) -> str:
    return f"0x{compiler_data.blueprint_bytecode.hex()}"


# EIP-170. Ref: https://eips.ethereum.org/EIPS/eip-170
EIP170_CONTRACT_SIZE_LIMIT: int = 2**14 + 2**13


def build_bytecode_runtime_output(compiler_data: CompilerData) -> str:
    compiled_bytecode_runtime_length = len(compiler_data.bytecode_runtime)
    if compiled_bytecode_runtime_length > EIP170_CONTRACT_SIZE_LIMIT:
        warnings.warn(
            f"Length of compiled bytecode is bigger than Ethereum contract size limit "
            "(see EIP-170: https://eips.ethereum.org/EIPS/eip-170): "
            f"{compiled_bytecode_runtime_length}b > {EIP170_CONTRACT_SIZE_LIMIT}b",
            ContractSizeLimitWarning,
            stacklevel=2,
        )
    return f"0x{compiler_data.bytecode_runtime.hex()}"


def build_opcodes_output(compiler_data: CompilerData) -> str:
    return _build_opcodes(compiler_data.bytecode)


def build_opcodes_runtime_output(compiler_data: CompilerData) -> str:
    return _build_opcodes(compiler_data.bytecode_runtime)


def _build_opcodes(bytecode: bytes) -> str:
    bytecode_sequence = deque(bytecode)

    opcode_map = dict((v[0], k) for k, v in opcodes.get_opcodes().items())
    opcode_output = []

    while bytecode_sequence:
        op = bytecode_sequence.popleft()
        opcode_output.append(opcode_map.get(op, f"VERBATIM_{hex(op)}"))
        if "PUSH" in opcode_output[-1] and opcode_output[-1] != "PUSH0":
            push_len = int(opcode_map[op][4:])
            # we can have push_len > len(bytecode_sequence) when there is data
            # (instead of code) at end of contract
            # CMC 2023-07-13 maybe just strip known data segments?
            push_len = min(push_len, len(bytecode_sequence))
            push_values = [f"{bytecode_sequence.popleft():0>2X}" for i in range(push_len)]
            opcode_output.append(f"0x{''.join(push_values)}")

    return " ".join(opcode_output)



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import NamespaceCollision, StructureException

fail_list = [  # noqa: E122
    """
@external
def foo(i: int128) -> int128:
    varő : int128 = i
    return varő
    """,
    """
@external
def foo(i: int128) -> int128:
    wei : int128 = i
    return wei
    """,
    """
@external
def foo(i: int128) -> int128:
    false : int128 = i
    return false
    """,
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_varname_validity_fail(bad_code):
    with raises(StructureException):
        compiler.compile_code(bad_code)


collision_fail_list = [
    """
@external
def foo(i: int128) -> int128:
    int128 : int128 = i
    return int128
    """,
    """
@external
def foo(i: int128) -> int128:
    decimal : int128 = i
    return decimal
    """,
]


@pytest.mark.parametrize("bad_code", collision_fail_list)
def test_varname_collision_fail(bad_code):
    with raises(NamespaceCollision):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo(i: int128) -> int128:
    variable : int128 = i
    return variable
    """,
    """
@external
def foo(i: int128) -> int128:
    var_123 : int128 = i
    return var_123
    """,
    """
@external
def foo(i: int128) -> int128:
    _var123 : int128 = i
    return _var123
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_varname_validity_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from tests.utils import ZERO_ADDRESS
from vyper.compiler import compile_code
from vyper.exceptions import EvmVersionException, VyperException

pytestmark = pytest.mark.requires_evm_version("cancun")


def test_transient_compiles():
    getter_code = """
my_map: public(transient(HashMap[address, uint256]))
    """
    t = compile_code(getter_code, output_formats=["opcodes_runtime"])
    t = t["opcodes_runtime"].split(" ")

    assert "TLOAD" in t
    assert "TSTORE" not in t

    setter_code = """
my_map: transient(HashMap[address, uint256])

@external
def setter(k: address, v: uint256):
    self.my_map[k] = v
    """
    t = compile_code(setter_code, output_formats=["opcodes_runtime"])
    t = t["opcodes_runtime"].split(" ")

    assert "TLOAD" not in t
    assert "TSTORE" in t

    getter_setter_code = """
my_map: public(transient(HashMap[address, uint256]))

@external
def setter(k: address, v: uint256):
    self.my_map[k] = v
    """
    t = compile_code(getter_setter_code, output_formats=["opcodes_runtime"])
    t = t["opcodes_runtime"].split(" ")

    assert "TLOAD" in t
    assert "TSTORE" in t


@pytest.mark.parametrize(
    "typ,value,zero",
    [
        ("uint256", 42, 0),
        ("int256", -(2**200), 0),
        ("int128", -(2**126), 0),
        ("address", "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", ZERO_ADDRESS),
        ("bytes32", b"deadbeef" * 4, b"\x00" * 32),
        ("bool", True, False),
        ("String[10]", "Vyper hiss", ""),
        ("Bytes[10]", b"Vyper hiss", b""),
    ],
)
def test_value_storage_retrieval(typ, value, zero, get_contract, env):
    code = f"""
bar: public(transient({typ}))

@external
def foo(a: {typ}) -> {typ}:
    self.bar = a
    return self.bar
    """

    c = get_contract(code)
    assert c.bar() == zero
    assert c.foo(value) == value
    env.clear_transient_storage()

    assert c.bar() == zero
    assert c.foo(value) == value
    env.clear_transient_storage()

    assert c.bar() == zero


@pytest.mark.parametrize("val", [0, 1, 2**256 - 1])
def test_usage_in_constructor(get_contract, val, env):
    code = """
A: transient(uint256)
a: public(uint256)


@deploy
def __init__(_a: uint256):
    self.A = _a
    self.a = self.A


@external
@view
def a1() -> uint256:
    return self.A
    """

    c = get_contract(code, val)
    assert c.a() == val
    env.clear_transient_storage()

    assert c.a1() == 0


def test_multiple_transient_values(get_contract, env):
    code = """
a: public(transient(uint256))
b: public(transient(address))
c: public(transient(String[64]))

@external
def foo(_a: uint256, _b: address, _c: String[64]) -> (uint256, address, String[64]):
    self.a = _a
    self.b = _b
    self.c = _c
    return self.a, self.b, self.c
    """

    try:
        compile_code(code)
    except VyperException as e:
        assert e.message.count("EvmVersionException") == 3
        # raise EvmVersionException to satisfy `requires_evm_version` behavior
        raise EvmVersionException()

    values = (3, "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", "Hello world")
    c = get_contract(code)
    assert c.foo(*values) == values
    env.clear_transient_storage()

    assert c.a() == 0
    assert c.b() == ZERO_ADDRESS
    assert c.c() == ""
    assert c.foo(*values) == values


def test_struct_transient(get_contract, env):
    code = """
struct MyStruct:
    a: uint256
    b: uint256
    c: address
    d: int256

my_struct: public(transient(MyStruct))

@external
def foo(_a: uint256, _b: uint256, _c: address, _d: int256) -> MyStruct:
    self.my_struct = MyStruct(
        a=_a,
        b=_b,
        c=_c,
        d=_d
    )
    return self.my_struct
    """
    values = (100, 42, "0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", -(2**200))

    c = get_contract(code)
    assert c.foo(*values) == values
    env.clear_transient_storage()

    assert c.my_struct() == (0, 0, ZERO_ADDRESS, 0)
    assert c.foo(*values) == values


def test_complex_struct_transient(get_contract, env):
    code = """
struct MyStruct:
    a: address
    b: MyStruct2
    c: DynArray[DynArray[uint256, 3], 3]

struct MyStruct2:
    a: DynArray[uint256, 3]

my_struct: public(transient(MyStruct))

@external
def foo(_a: address, _b: MyStruct2, _c: DynArray[DynArray[uint256, 3], 3]) -> MyStruct:
    self.my_struct = MyStruct(
        a=_a,
        b=_b,
        c=_c,
    )
    return self.my_struct
    """
    values = ("0xEeeeeEeeeEeEeeEeEeEeeEEEeeeeEeeeeeeeEEeE", ([1],), [[3, 4], [5, 6]])

    c = get_contract(code)
    assert c.foo(*values) == values
    env.clear_transient_storage()

    assert c.my_struct() == (ZERO_ADDRESS, ([],), [])
    assert c.foo(*values) == values


def test_complex_transient_modifiable(get_contract, env):
    code = """
struct MyStruct:
    a: uint256

my_struct: public(transient(MyStruct))

@external
def foo(a: uint256) -> MyStruct:
    self.my_struct = MyStruct(a=a)

    # struct members are modifiable after initialization
    self.my_struct.a += 1

    return self.my_struct
    """

    c = get_contract(code)
    assert c.foo(1) == (2,)
    env.clear_transient_storage()

    assert c.my_struct() == (0,)
    assert c.foo(1) == (2,)


def test_list_transient(get_contract, env):
    code = """
my_list: public(transient(uint256[3]))

@external
def foo(_a: uint256, _b: uint256, _c: uint256) -> uint256[3]:
    self.my_list = [_a, _b, _c]
    return self.my_list
    """
    values = (100, 42, 23230)

    c = get_contract(code)
    assert c.foo(*values) == list(values)
    env.clear_transient_storage()

    for i in range(3):
        assert c.my_list(i) == 0
    assert c.foo(*values) == list(values)


def test_hashmap_transient(get_contract, env):
    code = """
my_map: public(transient(HashMap[uint256, uint256]))

@external
def foo(k: uint256, v: uint256) -> uint256:
    self.my_map[k] = v
    return self.my_map[k]
    """
    c = get_contract(code)
    for v in range(5):
        for k in range(5):
            assert c.foo(k, v) == v
            env.clear_transient_storage()
            assert c.my_map(k) == 0


def test_complex_hashmap_transient(get_contract, env):
    code = """
struct MyStruct:
    a: uint256
    b: DynArray[uint256, 3]

my_map: public(transient(HashMap[uint256, MyStruct]))
my_res: public(HashMap[uint256, MyStruct])

@external
def do_side_effects():
    a: DynArray[uint256, 3] = [1, 2, 3]
    s: MyStruct = MyStruct(a=100, b=a)
    for i: uint256 in range(2):
        for j: uint256 in range(3):
            s.b[j] = i + j
        s.a = i
        self.my_map[i] = s
        self.my_res[i] = self.my_map[i]
    """
    c = get_contract(code)
    c.do_side_effects()
    for i in range(2):
        assert c.my_res(i)[0] == i
        assert c.my_map(i)[0] == 0
        env.clear_transient_storage()

        for j in range(3):
            assert c.my_res(i)[1][j] == i + j
            assert c.my_map(i)[1] == []


def test_dynarray_transient(get_contract, tx_failed, env):
    code = """
my_list: public(transient(DynArray[uint256, 3]))

@external
def get_my_list(_a: uint256, _b: uint256, _c: uint256) -> DynArray[uint256, 3]:
    self.my_list = [_a, _b, _c]
    return self.my_list

@external
def get_idx_two(_a: uint256, _b: uint256, _c: uint256) -> uint256:
    self.my_list = [_a, _b, _c]
    return self.my_list[2]
    """
    values = (100, 42, 23230)

    c = get_contract(code)
    assert c.get_my_list(*values) == list(values)
    env.clear_transient_storage()

    with tx_failed():
        c.my_list(0)
    assert c.get_idx_two(*values) == values[2]
    env.clear_transient_storage()

    with tx_failed():
        c.my_list(0)


def test_nested_dynarray_transient_2(get_contract):
    code = """
my_list: public(transient(DynArray[DynArray[uint256, 3], 3]))

@external
def get_my_list(_a: uint256, _b: uint256, _c: uint256) -> DynArray[DynArray[uint256, 3], 3]:
    self.my_list = [[_a, _b, _c], [_b, _a, _c], [_c, _b, _a]]
    return self.my_list

@external
def get_idx_two(_a: uint256, _b: uint256, _c: uint256) -> uint256:
    self.my_list = [[_a, _b, _c], [_b, _a, _c], [_c, _b, _a]]
    return self.my_list[2][2]
    """
    values = (100, 42, 23230)
    expected_values = [[100, 42, 23230], [42, 100, 23230], [23230, 42, 100]]

    c = get_contract(code)
    assert c.get_my_list(*values) == expected_values
    assert c.get_idx_two(*values) == expected_values[2][2]


def test_nested_dynarray_transient(get_contract, tx_failed, env):
    set_list = """
    self.my_list = [
        [[x, y, z], [y, z, x], [z, y, x]],
        [
            [x * 1000 + y, y * 1000 + z, z * 1000 + x],
            [- (x * 1000 + y), - (y * 1000 + z), - (z * 1000 + x)],
            [- (x * 1000) + y, - (y * 1000) + z, - (z * 1000) + x],
        ],
        [
            [z * 2, y * 3, x * 4],
            [z * (-2), y * (-3), x * (-4)],
            [z * (-y), y * (-x), x * (-z)],
        ],
    ]
    """
    code = f"""
interface Iface:
    def my_list(x: uint256, y: uint256, z: uint256) -> int128: view

my_list: public(transient(DynArray[DynArray[DynArray[int128, 3], 3], 3]))

@external
def get_my_list(x: int128, y: int128, z: int128) -> DynArray[DynArray[DynArray[int128, 3], 3], 3]:
    {set_list}
    return self.my_list

@external
def get_idx_two(x: int128, y: int128, z: int128) -> int128:
    {set_list}
    return self.my_list[2][2][2]

@external
def get_idx_two_using_getter(x: int128, y: int128, z: int128) -> int128:
    {set_list}
    return staticcall Iface(self).my_list(2, 2, 2)
    """
    values = (37, 41, 73)
    expected_values = [
        [[37, 41, 73], [41, 73, 37], [73, 41, 37]],
        [[37041, 41073, 73037], [-37041, -41073, -73037], [-36959, -40927, -72963]],
        [[146, 123, 148], [-146, -123, -148], [-2993, -1517, -2701]],
    ]

    c = get_contract(code)
    assert c.get_my_list(*values) == expected_values
    env.clear_transient_storage()

    with tx_failed():
        c.my_list(0, 0, 0)
    assert c.get_idx_two(*values) == expected_values[2][2][2]
    env.clear_transient_storage()

    with tx_failed():
        c.my_list(0, 0, 0)
    assert c.get_idx_two_using_getter(*values) == expected_values[2][2][2]
    env.clear_transient_storage()

    with tx_failed():
        c.my_list(0, 0, 0)


@pytest.mark.parametrize("n", range(5))
def test_internal_function_with_transient(get_contract, n, env):
    code = """
@internal
def foo() -> uint256:
    self.counter += 1
    return self.counter

counter: uint256
val: public(transient(uint256))

@external
def bar(x: uint256) -> uint256:
    self.counter = x
    self.foo()
    self.val = self.foo()
    return self.val
    """

    c = get_contract(code)
    assert c.bar(n) == n + 2
    env.clear_transient_storage()

    assert c.val() == 0
    assert c.bar(n) == n + 2


def test_nested_internal_function_transient(get_contract, env):
    code = """
d: public(uint256)
x: public(transient(uint256))

@deploy
def __init__():
    self.d = 1
    self.x = 2
    self.a()

@internal
def a():
    self.b()

@internal
def b():
    self.d = self.x
    """

    c = get_contract(code)
    assert c.d() == 2
    env.clear_transient_storage()

    assert c.x() == 0


def test_view_function_transient(get_contract, env):
    c1 = """
x: transient(uint256)

@external
def set_x(i: uint256):
    self.x = i

@external
@view
def get_x() -> uint256:
    return self.x
    """

    c2 = """
interface Foo:
    def set_x(i: uint256): nonpayable
    def get_x() -> uint256: view

@external
def bar(i: uint256, a: address) -> uint256:
    f: Foo = Foo(a)
    extcall f.set_x(i)
    return staticcall f.get_x()
    """

    c1 = get_contract(c1)
    c2 = get_contract(c2)

    value = 333
    assert c2.bar(value, c1.address) == value
    env.clear_transient_storage()

    assert c1.get_x() == 0


def test_modules_transient(get_contract, make_input_bundle):
    lib1 = """
counter: transient(uint256)
    """
    lib2 = """
import lib1

uses: lib1

counter: transient(uint256)
counter2: public(uint256)

@internal
def foo():
    lib1.counter += 1
    """
    main = """
import lib2
import lib1

initializes: lib2[lib1 := lib1]
initializes: lib1

@external
def foo() -> (uint256, uint256):
    lib1.counter = 2
    lib2.foo()
    lib2.counter = 10
    return lib1.counter, lib2.counter
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == (3, 10)


def test_complex_modules_transient(get_contract, make_input_bundle):
    lib1 = """
l: transient(uint256[3])
    """
    lib2 = """
import lib1

uses: lib1

struct MyStruct:
    a: uint256
    b: uint256

s: transient(MyStruct)

@internal
def foo():
    self.s = MyStruct(a=lib1.l[0], b=lib1.l[1])
    """
    main = """
import lib2
import lib1

initializes: lib2[lib1 := lib1]
initializes: lib1

my_map: HashMap[uint256, uint256]

@external
def foo() -> (uint256[3], uint256, uint256, uint256):
    lib1.l = [1, 2, 3]
    lib2.foo()
    self.my_map[0] = 42
    return lib1.l, lib2.s.a, lib2.s.b, self.my_map[0]
    """
    input_bundle = make_input_bundle({"lib1.vy": lib1, "lib2.vy": lib2})

    c = get_contract(main, input_bundle=input_bundle)
    assert c.foo() == ([1, 2, 3], 1, 2, 42)



# Test for Safe Remote Purchase
# (https://github.com/ethereum/solidity/blob/develop/docs/solidity-by-example.rst)
# ported to vyper and optimized

# Rundown of the transaction:
# 1. Seller posts item for sale and posts safety deposit of double the item
#    value. Balance is 2*value.
# (1.1. Seller can reclaim deposit and close the sale as long as nothing was purchased.)
# 2. Buyer purchases item (value) plus posts an additional safety deposit (Item
#    value). Balance is 4*value
# 3. Seller ships item
# 4. Buyer confirms receiving the item. Buyer's deposit (value) is returned.
#    Seller's deposit (2*value) + items value is returned. Balance is 0.
import pytest
from eth_utils import to_wei


@pytest.fixture(scope="module")
def contract_code(get_contract):
    with open("examples/safe_remote_purchase/safe_remote_purchase.vy") as f:
        contract_code = f.read()
    return contract_code


@pytest.fixture(scope="module")
def get_balance(env):
    def get_balance():
        a0, a1 = env.accounts[:2]
        # balance of a1 = seller, a2 = buyer
        return env.get_balance(a0), env.get_balance(a1)

    return get_balance


def test_initial_state(env, tx_failed, get_contract, get_balance, contract_code):
    env.set_balance(env.deployer, to_wei(2, "ether"))
    # Initial deposit has to be divisible by two
    with tx_failed():
        get_contract(contract_code, value=13)
    # Seller puts item up for sale
    a0_pre_bal, a1_pre_bal = get_balance()
    c = get_contract(contract_code, value=to_wei(2, "ether"))
    # Check that the seller is set correctly
    assert c.seller() == env.accounts[0]
    # Check if item value is set correctly (Half of deposit)
    assert c.value() == to_wei(1, "ether")
    # Check if unlocked() works correctly after initialization
    assert c.unlocked() is True
    # Check that sellers (and buyers) balance is correct
    assert get_balance() == ((a0_pre_bal - to_wei(2, "ether")), a1_pre_bal)


def test_abort(env, tx_failed, get_balance, get_contract, contract_code):
    a0, a1, a2 = env.accounts[:3]
    for a in env.accounts[:3]:
        env.set_balance(a, 10**20)

    a0_pre_bal, a1_pre_bal = get_balance()
    c = get_contract(contract_code, value=to_wei(2, "ether"))
    assert c.value() == to_wei(1, "ether")
    # Only sender can trigger refund
    with tx_failed():
        c.abort(sender=a2)
    # Refund works correctly
    c.abort(sender=a0)
    assert get_balance() == (a0_pre_bal, a1_pre_bal)
    # Purchase in process, no refund possible
    c = get_contract(contract_code, value=2)
    c.purchase(value=2, sender=a1)
    with tx_failed():
        c.abort(sender=a0)


def test_purchase(env, get_contract, tx_failed, get_balance, contract_code):
    a0, a1, a2, a3 = env.accounts[:4]
    for a in env.accounts[:4]:
        env.set_balance(a, 10**18)

    init_bal_a0, init_bal_a1 = get_balance()
    c = get_contract(contract_code, value=2)
    # Purchase for too low/high price
    with tx_failed():
        c.purchase(value=1, sender=a1)
    with tx_failed():
        c.purchase(value=3, sender=a1)
    # Purchase for the correct price
    c.purchase(value=2, sender=a1)
    # Check if buyer is set correctly
    assert c.buyer() == a1
    # Check if contract is locked correctly
    assert c.unlocked() is False
    # Check balances, both deposits should have been deducted
    assert get_balance() == (init_bal_a0 - 2, init_bal_a1 - 2)
    # Allow nobody else to purchase
    with tx_failed():
        c.purchase(value=2, sender=a3)


def test_received(env, get_contract, tx_failed, get_balance, contract_code):
    a0, a1 = env.accounts[:2]
    env.set_balance(a0, 10**18)
    env.set_balance(a1, 10**18)
    init_bal_a0, init_bal_a1 = get_balance()
    c = get_contract(contract_code, value=2)
    # Can only be called after purchase
    with tx_failed():
        c.received(sender=a1)
    # Purchase completed
    c.purchase(value=2, sender=a1)
    # Check that e.g. sender cannot trigger received
    with tx_failed():
        c.received(sender=a0)
    # Check if buyer can call receive
    c.received(sender=a1)
    # Final check if everything worked. 1 value has been transferred
    assert get_balance() == (init_bal_a0 + 1, init_bal_a1 - 1)


def test_received_reentrancy(env, get_contract, tx_failed, get_balance, contract_code):
    buyer_contract_code = """
interface PurchaseContract:

    def received(): nonpayable
    def purchase(): payable
    def unlocked() -> bool: view

purchase_contract: PurchaseContract


@deploy
def __init__(_purchase_contract: address):
    self.purchase_contract = PurchaseContract(_purchase_contract)


@payable
@external
def start_purchase():
    extcall self.purchase_contract.purchase(value=2)


@payable
@external
def start_received():
   extcall self.purchase_contract.received()


@external
@payable
def __default__():
    extcall self.purchase_contract.received()

    """

    a0, a1 = env.accounts[:2]
    for a in env.accounts[:2]:
        env.set_balance(a, 10**18)

    c = get_contract(contract_code, value=2)
    buyer_contract = get_contract(buyer_contract_code, *[c.address])
    buyer_contract_address = buyer_contract.address
    init_bal_a0, init_bal_buyer_contract = (
        env.get_balance(a0),
        env.get_balance(buyer_contract_address),
    )
    # Start purchase
    buyer_contract.start_purchase(value=4, sender=a1, gas=100000)
    assert c.unlocked() is False
    assert c.buyer() == buyer_contract_address

    # Trigger "re-entry"
    with tx_failed():
        buyer_contract.start_received(sender=a1, gas=100000)

    # Final check if everything worked. 1 value has been transferred
    assert env.get_balance(a0), env.get_balance(buyer_contract_address) == (
        init_bal_a0 + 1,
        init_bal_buyer_contract - 1,
    )



import decimal
import functools
import math

from vyper import ast as vy_ast
from vyper.codegen.core import (
    LOAD,
    IRnode,
    bytes_clamp,
    bytes_data_ptr,
    clamp,
    clamp_basetype,
    clamp_le,
    get_bytearray_length,
    int_clamp,
    is_bytes_m_type,
    is_decimal_type,
    is_flag_type,
    is_integer_type,
    sar,
    shl,
    shr,
    unwrap_location,
)
from vyper.codegen.expr import Expr
from vyper.exceptions import (
    CompilerPanic,
    InvalidLiteral,
    InvalidType,
    StructureException,
    TypeMismatch,
)
from vyper.semantics.types import (
    AddressT,
    BoolT,
    BytesM_T,
    BytesT,
    DecimalT,
    FlagT,
    IntegerT,
    StringT,
)
from vyper.semantics.types.bytestrings import _BytestringT
from vyper.semantics.types.shortcuts import INT256_T, UINT160_T, UINT256_T
from vyper.utils import DECIMAL_DIVISOR, round_towards_zero, unsigned_to_signed


def _FAIL(ityp, otyp, source_expr=None):
    # TODO consider changing this to InvalidType to be consistent
    # with the case when types are equal.
    raise TypeMismatch(f"Can't convert {ityp} to {otyp}", source_expr)


def _input_types(*allowed_types):
    def decorator(f):
        @functools.wraps(f)
        def check_input_type(expr, arg, out_typ):
            # convert arg to out_typ.
            # (expr is the AST corresponding to `arg`)
            ok = isinstance(arg.typ, allowed_types)
            if not ok:
                _FAIL(arg.typ, out_typ, expr)

            # user safety: disallow convert from type to itself
            # note allowance of [u]int256; this is due to type inference
            # on literals not quite working yet.
            if arg.typ == out_typ and arg.typ not in (UINT256_T, INT256_T):
                raise InvalidType(f"value and target are both {out_typ}", expr)

            return f(expr, arg, out_typ)

        return check_input_type

    return decorator


def _bytes_to_num(arg, out_typ, signed):
    # converting a bytestring to a number:
    # bytestring and bytes_m are right-padded with zeroes, int is left-padded.
    # convert by shr or sar the number of zero bytes (converted to bits)
    # e.g. "abcd000000000000" -> bitcast(000000000000abcd, output_type)

    if isinstance(arg.typ, _BytestringT):
        _len = get_bytearray_length(arg)
        arg = LOAD(bytes_data_ptr(arg))
        num_zero_bits = ["mul", 8, ["sub", 32, _len]]
    elif is_bytes_m_type(arg.typ):
        num_zero_bits = 8 * (32 - arg.typ.m)
    else:  # pragma: nocover
        raise CompilerPanic("unreachable")

    if signed:
        ret = sar(num_zero_bits, arg)
    else:
        ret = shr(num_zero_bits, arg)

    annotation = (f"__intrinsic__byte_array_to_num({out_typ})",)
    return IRnode.from_list(ret, annotation=annotation)


def _clamp_numeric_convert(arg, arg_bounds, out_bounds, arg_is_signed):
    arg_lo, arg_hi = arg_bounds
    out_lo, out_hi = out_bounds

    if arg_lo < out_lo:
        # if not arg_is_signed, arg_lo is 0, so this branch cannot be hit
        assert arg_is_signed, "bad assumption in numeric convert"
        arg = clamp("sge", arg, out_lo)

    if arg_hi > out_hi:
        # out_hi must be smaller than MAX_UINT256, so clample makes sense.
        # add an assertion, just in case this assumption ever changes.
        assert out_hi < 2**256 - 1, "bad assumption in numeric convert"
        arg = clamp_le(arg, out_hi, arg_is_signed)

    return arg


# truncate from fixed point decimal to int
def _fixed_to_int(arg, out_typ):
    DIVISOR = arg.typ.divisor

    # block inputs which are out of bounds before truncation.
    # e.g., convert(255.1, uint8) should revert or fail to compile.
    out_lo, out_hi = out_typ.int_bounds
    out_lo *= DIVISOR
    out_hi *= DIVISOR

    arg_bounds = arg.typ.int_bounds

    clamped_arg = _clamp_numeric_convert(arg, arg_bounds, (out_lo, out_hi), arg.typ.is_signed)

    assert arg.typ.is_signed, "should use unsigned div"  # stub for when we ufixed
    return IRnode.from_list(["sdiv", clamped_arg, DIVISOR], typ=out_typ)


# promote from int to fixed point decimal
def _int_to_fixed(arg, out_typ):
    DIVISOR = out_typ.divisor

    # block inputs which are out of bounds before promotion
    out_lo, out_hi = out_typ.int_bounds
    out_lo = round_towards_zero(out_lo / decimal.Decimal(DIVISOR))
    out_hi = round_towards_zero(out_hi / decimal.Decimal(DIVISOR))

    arg_bounds = arg.typ.int_bounds

    clamped_arg = _clamp_numeric_convert(arg, arg_bounds, (out_lo, out_hi), arg.typ.is_signed)

    return IRnode.from_list(["mul", clamped_arg, DIVISOR], typ=out_typ)


# clamp for dealing with conversions between int types (from arg to dst)
def _int_to_int(arg, out_typ):
    # do the same thing as
    # _clamp_numeric_convert(arg, arg.typ.int_bounds, out_typ.int_bounds, arg.typ.is_signed)
    # but with better code size and gas.
    if arg.typ.is_signed and not out_typ.is_signed:
        # e.g. (clample (clampge arg 0) (2**128 - 1))

        # note that when out_typ.bits == 256,
        # (clample arg 2**256 - 1) does not make sense.
        # see similar assertion in _clamp_numeric_convert.

        if out_typ.bits < arg.typ.bits:
            assert out_typ.bits < 256, "unreachable"
            # note: because of the usage of signed=False, and the fact
            # that out_bits < 256 in this branch, below implies
            # not only (clample arg 2**128 - 1) but also (clampge arg 0).
            arg = int_clamp(arg, out_typ.bits, signed=False)

        else:
            # note: this also works for out_bits == 256.
            arg = clamp("sge", arg, 0)

    elif not arg.typ.is_signed and out_typ.is_signed:
        # e.g. (uclample (uclampge arg 0) (2**127 - 1))
        # (note that (uclampge arg 0) always evaluates to true.)
        arg = int_clamp(arg, out_typ.bits - 1, signed=False)

    elif out_typ.bits < arg.typ.bits:
        assert out_typ.bits < 256, "unreachable"
        # narrowing conversion, signs are the same.
        # we can just use regular int clampers.
        arg = int_clamp(arg, out_typ.bits, out_typ.is_signed)

    else:
        # widening conversion, signs are the same.
        # we do not have to do any clamps.
        assert arg.typ.is_signed == out_typ.is_signed and out_typ.bits >= arg.typ.bits

    return IRnode.from_list(arg, typ=out_typ)


def _check_bytes(expr, arg, output_type, max_bytes_allowed):
    if isinstance(arg.typ, _BytestringT):
        if arg.typ.maxlen > max_bytes_allowed:
            _FAIL(arg.typ, output_type, expr)
    else:
        # sanity check. should not have conversions to non-base types
        assert output_type.memory_bytes_required == 32


# apply sign extension, if expected. note that the sign bit
# is always taken to be the first bit of the bytestring.
# (e.g. convert(0xff <bytes1>, int16) == -1)
def _signextend(expr, val, arg_typ):
    if isinstance(expr, vy_ast.Hex):
        assert len(expr.value[2:]) // 2 == arg_typ.m
        n_bits = arg_typ.m_bits
    else:
        assert len(expr.value) == arg_typ.maxlen
        n_bits = arg_typ.maxlen * 8

    return unsigned_to_signed(val, n_bits)


def _literal_int(expr, arg_typ, out_typ):
    # TODO: possible to reuse machinery from expr.py?
    if isinstance(expr, vy_ast.Hex):
        val = int(expr.value, 16)
    elif isinstance(expr, vy_ast.Bytes):
        val = int.from_bytes(expr.value, "big")
    elif isinstance(expr, (vy_ast.Int, vy_ast.Decimal, vy_ast.NameConstant)):
        val = expr.value
    else:  # pragma: no cover
        raise CompilerPanic("unreachable")

    if isinstance(expr, (vy_ast.Hex, vy_ast.Bytes)) and out_typ.is_signed:
        val = _signextend(expr, val, arg_typ)

    lo, hi = out_typ.int_bounds
    if not (lo <= val <= hi):
        raise InvalidLiteral("Number out of range", expr)

    # cast to int AFTER bounds check (ensures decimal is in bounds before truncation)
    val = int(val)

    return IRnode.from_list(val, typ=out_typ)


def _literal_decimal(expr, arg_typ, out_typ):
    if isinstance(expr, vy_ast.Hex):
        val = decimal.Decimal(int(expr.value, 16))
    else:
        val = decimal.Decimal(expr.value)  # should work for Int, Decimal
        val *= DECIMAL_DIVISOR

    # sanity check type checker did its job
    assert math.ceil(val) == math.floor(val)

    val = int(val)

    # apply sign extension, if expected
    if isinstance(expr, (vy_ast.Hex, vy_ast.Bytes)) and out_typ.is_signed:
        val = _signextend(expr, val, arg_typ)

    lo, hi = out_typ.int_bounds
    if not lo <= val <= hi:
        raise InvalidLiteral("Number out of range", expr)

    return IRnode.from_list(val, typ=out_typ)


# any base type or bytes/string
@_input_types(IntegerT, DecimalT, BytesM_T, AddressT, BoolT, BytesT, StringT)
def to_bool(expr, arg, out_typ):
    _check_bytes(expr, arg, out_typ, 32)  # should we restrict to Bytes[1]?

    if isinstance(arg.typ, _BytestringT):
        # no clamp. checks for any nonzero bytes.
        arg = _bytes_to_num(arg, out_typ, signed=False)

    # NOTE: for decimal, the behavior is x != 0.0,
    # (we do not issue an `sdiv DECIMAL_DIVISOR`)

    return IRnode.from_list(["iszero", ["iszero", arg]], typ=out_typ)


@_input_types(IntegerT, DecimalT, BytesM_T, AddressT, BoolT, FlagT, BytesT)
def to_int(expr, arg, out_typ):
    return _to_int(expr, arg, out_typ)


# an internal version of to_int without input validation
def _to_int(expr, arg, out_typ):
    assert out_typ.bits % 8 == 0
    _check_bytes(expr, arg, out_typ, 32)

    if isinstance(expr, vy_ast.Constant):
        return _literal_int(expr, arg.typ, out_typ)

    elif isinstance(arg.typ, BytesT):
        arg_typ = arg.typ
        arg = _bytes_to_num(arg, out_typ, signed=out_typ.is_signed)
        if arg_typ.maxlen * 8 > out_typ.bits:
            arg = int_clamp(arg, out_typ.bits, signed=out_typ.is_signed)

    elif is_bytes_m_type(arg.typ):
        arg_typ = arg.typ
        arg = _bytes_to_num(arg, out_typ, signed=out_typ.is_signed)
        if arg_typ.m_bits > out_typ.bits:
            arg = int_clamp(arg, out_typ.bits, signed=out_typ.is_signed)

    elif is_decimal_type(arg.typ):
        arg = _fixed_to_int(arg, out_typ)

    elif is_flag_type(arg.typ):
        if out_typ != UINT256_T:
            _FAIL(arg.typ, out_typ, expr)
        # pretend flag is uint256
        arg = IRnode.from_list(arg, typ=UINT256_T)
        # use int_to_int rules
        arg = _int_to_int(arg, out_typ)

    elif is_integer_type(arg.typ):
        arg = _int_to_int(arg, out_typ)

    elif arg.typ == AddressT():
        if out_typ.is_signed:
            # TODO if possible, refactor to move this validation close to the entry of the function
            _FAIL(arg.typ, out_typ, expr)
        if out_typ.bits < 160:
            arg = int_clamp(arg, out_typ.bits, signed=False)

    return IRnode.from_list(arg, typ=out_typ)


@_input_types(IntegerT, BoolT, BytesM_T, BytesT)
def to_decimal(expr, arg, out_typ):
    _check_bytes(expr, arg, out_typ, 32)

    if isinstance(expr, vy_ast.Constant):
        return _literal_decimal(expr, arg.typ, out_typ)

    if isinstance(arg.typ, BytesT):
        arg_typ = arg.typ
        arg = _bytes_to_num(arg, out_typ, signed=True)
        if arg_typ.maxlen * 8 > 168:
            arg = IRnode.from_list(arg, typ=out_typ)
            arg = clamp_basetype(arg)

        return IRnode.from_list(arg, typ=out_typ)

    elif is_bytes_m_type(arg.typ):
        arg_typ = arg.typ
        arg = _bytes_to_num(arg, out_typ, signed=True)
        if arg_typ.m_bits > 168:
            arg = IRnode.from_list(arg, typ=out_typ)
            arg = clamp_basetype(arg)

        return IRnode.from_list(arg, typ=out_typ)

    elif is_integer_type(arg.typ):
        arg = _int_to_fixed(arg, out_typ)
        return IRnode.from_list(arg, typ=out_typ)

    elif arg.typ == BoolT():
        # TODO: consider adding is_signed and bits to bool so we can use _int_to_fixed
        arg = ["mul", arg, 10**out_typ.decimals]
        return IRnode.from_list(arg, typ=out_typ)
    else:  # pragma: nocover
        raise CompilerPanic("unreachable")


@_input_types(IntegerT, DecimalT, BytesM_T, AddressT, BytesT, BoolT)
def to_bytes_m(expr, arg, out_typ):
    _check_bytes(expr, arg, out_typ, max_bytes_allowed=out_typ.m)

    if isinstance(arg.typ, BytesT):
        bytes_val = LOAD(bytes_data_ptr(arg))

        # zero out any dirty bytes (which can happen in the last
        # word of a bytearray)
        len_ = get_bytearray_length(arg)
        num_zero_bits = IRnode.from_list(["mul", ["sub", 32, len_], 8])
        with num_zero_bits.cache_when_complex("bits") as (b, num_zero_bits):
            arg = shl(num_zero_bits, shr(num_zero_bits, bytes_val))
            arg = b.resolve(arg)

    elif is_bytes_m_type(arg.typ):
        # clamp if it's a downcast
        if arg.typ.m > out_typ.m:
            arg = bytes_clamp(arg, out_typ.m)

    elif is_integer_type(arg.typ) or arg.typ == AddressT():
        if arg.typ == AddressT():
            int_bits = 160
        else:
            int_bits = arg.typ.bits

        if out_typ.m_bits < int_bits:
            # question: allow with runtime clamp?
            # arg = int_clamp(m_bits, signed=arg.typ.signed)
            _FAIL(arg.typ, out_typ, expr)

        # note: neg numbers not OOB. keep sign bit
        arg = shl(256 - out_typ.m_bits, arg)

    elif is_decimal_type(arg.typ):
        if out_typ.m_bits < arg.typ.bits:
            _FAIL(arg.typ, out_typ, expr)

        # note: neg numbers not OOB. keep sign bit
        arg = shl(256 - out_typ.m_bits, arg)

    else:
        # bool
        arg = shl(256 - out_typ.m_bits, arg)

    return IRnode.from_list(arg, typ=out_typ)


@_input_types(BytesM_T, IntegerT, BytesT)
def to_address(expr, arg, out_typ):
    # question: should this be allowed?
    if is_integer_type(arg.typ):
        if arg.typ.is_signed:
            _FAIL(arg.typ, out_typ, expr)

    ret = _to_int(expr, arg, UINT160_T)
    return IRnode.from_list(ret, out_typ)


def _cast_bytestring(expr, arg, out_typ):
    # ban converting Bytes[20] to Bytes[21]
    if isinstance(arg.typ, out_typ.__class__) and arg.typ.maxlen <= out_typ.maxlen:
        _FAIL(arg.typ, out_typ, expr)
    # can't downcast literals with known length (e.g. b"abc" to Bytes[2])
    if isinstance(expr, vy_ast.Constant) and arg.typ.maxlen > out_typ.maxlen:
        _FAIL(arg.typ, out_typ, expr)

    ret = ["seq"]
    if out_typ.maxlen < arg.typ.maxlen:
        ret.append(["assert", ["le", get_bytearray_length(arg), out_typ.maxlen]])
    ret.append(arg)
    # NOTE: this is a pointer cast
    return IRnode.from_list(ret, typ=out_typ, location=arg.location, encoding=arg.encoding)


# question: should we allow bytesM -> String?
@_input_types(BytesT, StringT)
def to_string(expr, arg, out_typ):
    return _cast_bytestring(expr, arg, out_typ)


@_input_types(StringT, BytesT)
def to_bytes(expr, arg, out_typ):
    return _cast_bytestring(expr, arg, out_typ)


@_input_types(IntegerT)
def to_flag(expr, arg, out_typ):
    if arg.typ != UINT256_T:
        _FAIL(arg.typ, out_typ, expr)

    if len(out_typ._flag_members) < 256:
        arg = int_clamp(arg, bits=len(out_typ._flag_members), signed=False)

    return IRnode.from_list(arg, typ=out_typ)


def convert(expr, context):
    assert len(expr.args) == 2, "bad typecheck: convert"

    arg_ast = expr.args[0]
    arg = Expr(arg_ast, context).ir_node
    original_arg = arg

    out_typ = expr.args[1]._metadata["type"].typedef

    if arg.typ._is_prim_word:
        arg = unwrap_location(arg)
    with arg.cache_when_complex("arg") as (b, arg):
        if out_typ == BoolT():
            ret = to_bool(arg_ast, arg, out_typ)
        elif out_typ == AddressT():
            ret = to_address(arg_ast, arg, out_typ)
        elif is_flag_type(out_typ):
            ret = to_flag(arg_ast, arg, out_typ)
        elif is_integer_type(out_typ):
            ret = to_int(arg_ast, arg, out_typ)
        elif is_bytes_m_type(out_typ):
            ret = to_bytes_m(arg_ast, arg, out_typ)
        elif is_decimal_type(out_typ):
            ret = to_decimal(arg_ast, arg, out_typ)
        elif isinstance(out_typ, BytesT):
            ret = to_bytes(arg_ast, arg, out_typ)
        elif isinstance(out_typ, StringT):
            ret = to_string(arg_ast, arg, out_typ)
        else:
            raise StructureException(f"Conversion to {out_typ} is invalid.", arg_ast)

        # test if arg actually changed. if not, we do not need to use
        # unwrap_location (this can reduce memory traffic for downstream
        # operations which are in-place, like the returndata routine)
        test_arg = IRnode.from_list(arg, typ=out_typ)
        if test_arg == ret:
            original_arg.typ = out_typ
            return original_arg

        return IRnode.from_list(b.resolve(ret))



import pytest
from pytest import raises

from vyper import compiler
from vyper.exceptions import InvalidType, NamespaceCollision, StructureException, SyntaxException

fail_list = [  # noqa: E122
    (
        """
event Âssign:
    variable: int128

@external
def foo(i: int128) -> int128:
    temp_var : int128 = i
    log Âssign(temp_var)
    return temp_var
    """,
        StructureException,
    ),
    (
        """
event int128:
    variable: int128

@external
def foo(i: int128) -> int128:
    temp_var : int128 = i
    log int128(temp_var)
    return temp_var
    """,
        NamespaceCollision,
    ),
    (
        """
event decimal:
    variable: int128

@external
def foo(i: int128) -> int128:
    temp_var : int128 = i
    log decimal(temp_var)
    return temp_var
    """,
        NamespaceCollision,
    ),
    (
        """
event wei:
    variable: int128

@external
def foo(i: int128) -> int128:
    temp_var : int128 = i
    log wei(temp_var)
    return temp_var
    """,
        StructureException,
    ),
    (
        """
event false:
    variable: int128

@external
def foo(i: int128) -> int128:
    temp_var : int128 = i
    log false(temp_var)
    return temp_var
    """,
        StructureException,
    ),
    (
        """
Transfer: eve.t({_from: indexed(address)})
    """,
        SyntaxException,
    ),
    (
        """
event Transfer:
    _from: i.dexed(address)
    _to: indexed(address)
    lue: uint256
    """,
        InvalidType,
    ),
]


@pytest.mark.parametrize("bad_code,exc", fail_list)
def test_varname_validity_fail(bad_code, exc):
    with raises(exc):
        compiler.compile_code(bad_code)


valid_list = [
    """
event Assigned:
    variable: int128

@external
def foo(i: int128) -> int128:
    variable : int128 = i
    log Assigned(variable)
    return variable
    """,
    """
event _Assign:
    variable: int128

@external
def foo(i: int128) -> int128:
    variable : int128 = i
    log _Assign(variable)
    return variable
    """,
    """
event Assigned1:
    variable: int128

@external
def foo(i: int128) -> int128:
    variable : int128 = i
    log Assigned1(variable)
    return variable
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_varname_validity_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    """
@external
def foo():
    selfdestruct(7)
    """
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_block_fail(bad_code):
    with pytest.raises(TypeMismatch):
        compiler.compile_code(bad_code)


valid_list = [
    """
@external
def foo():
    selfdestruct(0x1234567890123456789012345678901234567890)
    """
]


@pytest.mark.parametrize("good_code", valid_list)
def test_block_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest


def test_string_return(get_contract):
    code = """
@external
def testb() -> String[100]:
    a: String[100] = "test return"
    return a

@external
def testa(inp: String[100]) -> String[100]:
    return inp
    """

    c = get_contract(code)

    assert c.testa("meh") == "meh"
    assert c.testb() == "test return"


def test_string_concat(get_contract):
    code = """
@external
def testb(inp: String[10]) -> String[128]:
    a: String[100] = "return message:"
    b: String[128] = concat(a, " ", inp)
    return b

@external
def testa(inp: String[10]) -> String[160]:
    a: String[100] = "<-- return message"
    return concat("Funny ", inp, " ", inp, a)
    """

    c = get_contract(code)

    assert c.testb("bob") == "return message: bob"
    assert c.testa("foo") == "Funny foo foo<-- return message"


def test_basic_long_string_as_keys(get_contract):
    code = """
mapped_string: HashMap[String[34], int128]

@external
def set(k: String[34], v: int128):
    self.mapped_string[k] = v

@external
def get(k: String[34]) -> int128:
    return self.mapped_string[k]
    """

    c = get_contract(code)

    c.set("a" * 34, 6789, gas=10**6)

    assert c.get("a" * 34) == 6789


def test_string_slice(get_contract, tx_failed):
    test_slice4 = """
@external
def foo(inp: String[10], start: uint256, _len: uint256) -> String[10]:
    return slice(inp, start, _len)
    """

    c = get_contract(test_slice4)
    assert c.foo("badminton", 3, 3) == "min"
    assert c.foo("badminton", 0, 9) == "badminton"
    assert c.foo("badminton", 1, 8) == "adminton"
    assert c.foo("badminton", 1, 7) == "adminto"
    assert c.foo("badminton", 1, 0) == ""
    assert c.foo("badminton", 9, 0) == ""

    with tx_failed():
        c.foo("badminton", 0, 10)
    with tx_failed():
        c.foo("badminton", 1, 9)
    with tx_failed():
        c.foo("badminton", 9, 1)
    with tx_failed():
        c.foo("badminton", 10, 0)


def test_private_string(get_contract):
    private_test_code = """
greeting: public(String[100])

@deploy
def __init__():
    self.greeting = "Hello "

@internal
def construct(greet: String[100]) -> String[200]:
    return concat(self.greeting, greet)

@external
def hithere(name: String[100]) -> String[200]:
    d: String[200] = self.construct(name)
    return d
    """

    c = get_contract(private_test_code)
    assert c.hithere("bob") == "Hello bob"
    assert c.hithere("alice") == "Hello alice"


def test_logging_extended_string(get_contract, get_logs):
    code = """
event MyLog:
    arg1: int128
    arg2: String[64]
    arg3: int128

@external
def foo():
    log MyLog(667788, 'hellohellohellohellohellohellohellohellohello', 334455)
    """

    c = get_contract(code)
    c.foo()
    (log,) = get_logs(c, "MyLog")

    assert log.args.arg1 == 667788
    assert log.args.arg2 == "hello" * 9
    assert log.args.arg3 == 334455


def test_tuple_return_external_contract_call_string(get_contract):
    contract_1 = """
@external
def out_literals() -> (int128, address, String[10]):
    return 1, 0x0000000000000000000000000000000000000123, "random"
    """

    contract_2 = """
interface Test:
    def out_literals() -> (int128, address, String[10]) : view

@external
def test(addr: address) -> (int128, address, String[10]):
    a: int128 = 0
    b: address = empty(address)
    c: String[10] = ""
    (a, b, c) = staticcall Test(addr).out_literals()
    return a, b,c
    """
    c1 = get_contract(contract_1)
    c2 = get_contract(contract_2)

    assert c1.out_literals() == (1, "0x0000000000000000000000000000000000000123", "random")
    assert c2.test(c1.address) == (1, "0x0000000000000000000000000000000000000123", "random")


def test_default_arg_string(get_contract):
    code = """
@external
def test(a: uint256, b: String[50] = "foo") -> Bytes[100]:
    return concat(
        convert(a, bytes32),
        convert(b, Bytes[50])
    )
    """

    c = get_contract(code)

    assert c.test(12345)[-3:] == b"foo"
    assert c.test(12345, "bar")[-3:] == b"bar"


string_equality_tests = [
    (
        100,
        "The quick brown fox jumps over the lazy dog",
        "The quick brown fox jumps over the lazy hog",
    ),
    # check <= 32 codepath
    (32, "abc", "abc\0"),
    (32, "abc", "abc\1"),  # use a_init dirty bytes
    (32, "abc\2", "abc"),  # use b_init dirty bytes
    (32, "", "\0"),
    (32, "", "\1"),
    (33, "", "\1"),
    (33, "", "\0"),
]


@pytest.mark.parametrize("len_,a,b", string_equality_tests)
def test_string_equality(get_contract, len_, a, b):
    # fixtures to initialize strings with dirty bytes
    a_init = "\\1" * len_
    b_init = "\\2" * len_
    string1 = a.encode("unicode_escape").decode("utf-8")
    string2 = b.encode("unicode_escape").decode("utf-8")
    code = f"""
a: String[{len_}]
b: String[{len_}]

@external
def equal_true() -> bool:
    a: String[{len_}] = "{a_init}"
    b: String[{len_}] = "{b_init}"
    a = "{string1}"
    b = "{string1}"
    return a == b

@external
def equal_false() -> bool:
    a: String[{len_}] = "{a_init}"
    b: String[{len_}] = "{b_init}"
    a = "{string1}"
    b = "{string2}"
    return a == b

@external
def not_equal_true() -> bool:
    a: String[{len_}] = "{a_init}"
    b: String[{len_}] = "{b_init}"
    a = "{string1}"
    b = "{string2}"
    return a != b

@external
def not_equal_false() -> bool:
    a: String[{len_}] = "{a_init}"
    b: String[{len_}] = "{b_init}"
    a = "{string1}"
    b = "{string1}"
    return a != b

@external
def literal_equal_true() -> bool:
    return "{string1}" == "{string1}"

@external
def literal_equal_false() -> bool:
    return "{string1}" == "{string2}"

@external
def literal_not_equal_true() -> bool:
    return "{string1}" != "{string2}"

@external
def literal_not_equal_false() -> bool:
    return "{string1}" != "{string1}"

@external
def storage_equal_true() -> bool:
    self.a = "{a_init}"
    self.b = "{b_init}"
    self.a = "{string1}"
    self.b = "{string1}"
    return self.a == self.b

@external
def storage_equal_false() -> bool:
    self.a = "{a_init}"
    self.b = "{b_init}"
    self.a = "{string1}"
    self.b = "{string2}"
    return self.a == self.b

@external
def storage_not_equal_true() -> bool:
    self.a = "{a_init}"
    self.b = "{b_init}"
    self.a = "{string1}"
    self.b = "{string2}"
    return self.a != self.b

@external
def storage_not_equal_false() -> bool:
    self.a = "{a_init}"
    self.b = "{b_init}"
    self.a = "{string1}"
    self.b = "{string1}"
    return self.a != self.b

@external
def string_compare_equal(str1: String[{len_}], str2: String[{len_}]) -> bool:
    return str1 == str2

@external
def string_compare_not_equal(str1: String[{len_}], str2: String[{len_}]) -> bool:
    return str1 != str2

@external
def compare_passed_storage_equal(str_: String[{len_}]) -> bool:
    self.a = "{a_init}"
    self.a = "{string1}"
    return self.a == str_

@external
def compare_passed_storage_not_equal(str_: String[{len_}]) -> bool:
    self.a = "{a_init}"
    self.a = "{string1}"
    return self.a != str_

@external
def compare_var_storage_equal_true() -> bool:
    self.a = "{a_init}"
    b: String[{len_}] = "{b_init}"
    self.a = "{string1}"
    b = "{string1}"
    return self.a == b

@external
def compare_var_storage_equal_false() -> bool:
    self.a = "{a_init}"
    b: String[{len_}] = "{b_init}"
    self.a = "{string1}"
    b = "{string2}"
    return self.a == b

@external
def compare_var_storage_not_equal_true() -> bool:
    self.a = "{a_init}"
    b: String[{len_}] = "{b_init}"
    self.a = "{string1}"
    b = "{string2}"
    return self.a != b

@external
def compare_var_storage_not_equal_false() -> bool:
    self.a = "{a_init}"
    b: String[{len_}] = "{b_init}"
    self.a = "{string1}"
    b = "{string1}"
    return self.a != b
    """

    c = get_contract(code)
    assert c.equal_true() is True
    assert c.equal_false() is False
    assert c.not_equal_true() is True
    assert c.not_equal_false() is False
    assert c.literal_equal_true() is True
    assert c.literal_equal_false() is False
    assert c.literal_not_equal_true() is True
    assert c.literal_not_equal_false() is False
    assert c.storage_equal_true() is True
    assert c.storage_equal_false() is False
    assert c.storage_not_equal_true() is True
    assert c.storage_not_equal_false() is False

    assert c.string_compare_equal(a, a) is True
    assert c.string_compare_equal(a, b) is False
    assert c.string_compare_not_equal(b, a) is True
    assert c.string_compare_not_equal(b, b) is False

    assert c.compare_passed_storage_equal(a) is True
    assert c.compare_passed_storage_equal(b) is False
    assert c.compare_passed_storage_not_equal(a) is False
    assert c.compare_passed_storage_not_equal(b) is True

    assert c.compare_var_storage_equal_true() is True
    assert c.compare_var_storage_equal_false() is False
    assert c.compare_var_storage_not_equal_true() is True
    assert c.compare_var_storage_not_equal_false() is False



def test_module_constant(make_input_bundle, get_contract):
    mod1 = """
X: constant(uint256) = 12345
    """
    contract = """
import mod1

@external
def foo() -> uint256:
    return mod1.X
    """

    input_bundle = make_input_bundle({"mod1.vy": mod1})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.foo() == 12345


def test_nested_module_constant(make_input_bundle, get_contract):
    # test nested module constants
    # test at least 3 modules deep to test the `path.reverse()` gizmo
    # in ConstantFolder.visit_Attribute()
    mod1 = """
X: constant(uint256) = 12345
    """
    mod2 = """
import mod1
X: constant(uint256) = 54321
    """
    mod3 = """
import mod2
X: constant(uint256) = 98765
    """

    contract = """
import mod1
import mod2
import mod3

@external
def test_foo() -> bool:
    assert mod1.X == 12345
    assert mod2.X == 54321
    assert mod3.X == 98765
    assert mod2.mod1.X == mod1.X
    assert mod3.mod2.mod1.X == mod1.X
    assert mod3.mod2.X == mod2.X
    return True
    """

    input_bundle = make_input_bundle({"mod1.vy": mod1, "mod2.vy": mod2, "mod3.vy": mod3})

    c = get_contract(contract, input_bundle=input_bundle)
    assert c.test_foo() is True


def test_import_constant_array(make_input_bundle, get_contract, tx_failed):
    mod1 = """
X: constant(uint256[3]) = [1,2,3]
    """
    contract = """
import mod1

@external
def foo(ix: uint256) -> uint256:
    return mod1.X[ix]
    """

    input_bundle = make_input_bundle({"mod1.vy": mod1})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.foo(0) == 1
    assert c.foo(1) == 2
    assert c.foo(2) == 3
    with tx_failed():
        c.foo(3)


def test_module_constant_builtin(make_input_bundle, get_contract):
    # test empty builtin, which is not (currently) foldable 2024-02-06
    mod1 = """
X: constant(uint256) = empty(uint256)
    """
    contract = """
import mod1

@external
def foo() -> uint256:
    return mod1.X
    """

    input_bundle = make_input_bundle({"mod1.vy": mod1})

    c = get_contract(contract, input_bundle=input_bundle)

    assert c.foo() == 0



from eth.codecs import abi

from vyper.utils import method_id


def test_revert_reason(env, tx_failed, get_contract):
    reverty_code = """
@external
def foo():
    data: Bytes[4] = method_id("NoFives()")
    raw_revert(data)
    """

    revert_bytes = method_id("NoFives()")

    with tx_failed(exc_text=revert_bytes.hex()):
        get_contract(reverty_code).foo()


def test_revert_reason_typed(env, tx_failed, get_contract):
    reverty_code = """
@external
def foo():
    val: uint256 = 5
    data: Bytes[100] = _abi_encode(val, method_id=method_id("NoFives(uint256)"))
    raw_revert(data)
    """

    revert_bytes = method_id("NoFives(uint256)") + abi.encode("(uint256)", (5,))

    with tx_failed(exc_text=revert_bytes.hex()):
        get_contract(reverty_code).foo()


def test_revert_reason_typed_no_variable(env, tx_failed, get_contract):
    reverty_code = """
@external
def foo():
    val: uint256 = 5
    raw_revert(_abi_encode(val, method_id=method_id("NoFives(uint256)")))
    """

    revert_bytes = method_id("NoFives(uint256)") + abi.encode("(uint256)", (5,))

    with tx_failed(exc_text=revert_bytes.hex()):
        get_contract(reverty_code).foo()



from decimal import Decimal

import pytest
from hypothesis import example, given, settings
from hypothesis import strategies as st

from tests.utils import decimal_to_int, parse_and_fold

st_decimals = st.decimals(
    min_value=-(2**32), max_value=2**32, allow_nan=False, allow_infinity=False, places=10
)


@pytest.mark.fuzzing
@settings(max_examples=50)
@given(value=st_decimals)
@example(value=Decimal("0.9999999999"))
@example(value=Decimal("0.0000000001"))
@example(value=Decimal("-0.9999999999"))
@example(value=Decimal("-0.0000000001"))
@pytest.mark.parametrize("fn_name", ["floor", "ceil"])
def test_floor_ceil(get_contract, value, fn_name):
    source = f"""
@external
def foo(a: decimal) -> int256:
    return {fn_name}(a)
    """
    contract = get_contract(source)

    vyper_ast = parse_and_fold(f"{fn_name}({value})")
    old_node = vyper_ast.body[0].value
    new_node = old_node.get_folded_value()

    assert isinstance(new_node.value, int)
    assert contract.foo(decimal_to_int(value)) == new_node.value



from typing import Optional

from vyper.venom.analysis.analysis import IRAnalysesCache, IRAnalysis
from vyper.venom.basicblock import IRInstruction, IRVariable
from vyper.venom.function import IRFunction


class DFGAnalysis(IRAnalysis):
    _dfg_inputs: dict[IRVariable, list[IRInstruction]]
    _dfg_outputs: dict[IRVariable, IRInstruction]

    def __init__(self, analyses_cache: IRAnalysesCache, function: IRFunction):
        super().__init__(analyses_cache, function)
        self._dfg_inputs = dict()
        self._dfg_outputs = dict()

    # return uses of a given variable
    def get_uses(self, op: IRVariable) -> list[IRInstruction]:
        return self._dfg_inputs.get(op, [])

    # the instruction which produces this variable.
    def get_producing_instruction(self, op: IRVariable) -> Optional[IRInstruction]:
        return self._dfg_outputs.get(op)

    @property
    def outputs(self) -> dict[IRVariable, IRInstruction]:
        return self._dfg_outputs

    def analyze(self):
        # Build DFG

        # %15 = add %13 %14
        # %16 = iszero %15
        # dfg_outputs of %15 is (%15 = add %13 %14)
        # dfg_inputs of %15 is all the instructions which *use* %15, ex. [(%16 = iszero %15), ...]
        for bb in self.function.get_basic_blocks():
            for inst in bb.instructions:
                operands = inst.get_inputs()
                res = inst.get_outputs()

                for op in operands:
                    inputs = self._dfg_inputs.setdefault(op, [])
                    inputs.append(inst)

                for op in res:  # type: ignore
                    self._dfg_outputs[op] = inst

    def as_graph(self) -> str:
        """
        Generate a graphviz representation of the dfg
        """
        lines = ["digraph dfg_graph {"]
        for var, inputs in self._dfg_inputs.items():
            for input in inputs:
                for op in input.get_outputs():
                    if isinstance(op, IRVariable):
                        lines.append(f'    " {var.name} " -> " {op.name} "')

        lines.append("}")
        return "\n".join(lines)

    def __repr__(self) -> str:
        return self.as_graph()



import re

import pytest

from vyper import compiler
from vyper.exceptions import ArgumentException, StructureException, TypeMismatch, UnknownType

fail_list = [
    (
        """
@external
def foo():
    for a[1]: uint256 in range(10):
        pass
    """,
        StructureException,
        "Invalid syntax for loop iterator",
        None,
        "a[1]",
    ),
    (
        """
@external
def bar():
    for i: uint256 in range(1,2,bound=0):
        pass
    """,
        StructureException,
        "Bound must be at least 1",
        None,
        "0",
    ),
    (
        """
@external
def foo():
    x: uint256 = 100
    for _: uint256 in range(10, bound=x):
        pass
    """,
        StructureException,
        "Bound must be a literal integer",
        None,
        "x",
    ),
    (
        """
@external
def foo():
    for _: uint256 in range(10, 20, bound=5):
        pass
    """,
        StructureException,
        "Please remove the `bound=` kwarg when using range with constants",
        None,
        "5",
    ),
    (
        """
@external
def foo():
    for _: uint256 in range(10, 20, bound=0):
        pass
    """,
        StructureException,
        "Bound must be at least 1",
        None,
        "0",
    ),
    (
        """
@external
def bar():
    x:uint256 = 1
    for i: uint256 in range(x,x+1,bound=2,extra=3):
        pass
    """,
        ArgumentException,
        "Invalid keyword argument 'extra'",
        None,
        "extra=3",
    ),
    (
        """
@external
def bar():
    for i: uint256 in range(0):
        pass
    """,
        StructureException,
        "End must be greater than start",
        None,
        "0",
    ),
    (
        """
@external
def bar():
    x:uint256 = 1
    for i: uint256 in range(x):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "x",
    ),
    (
        """
@external
def bar():
    x:uint256 = 1
    for i: uint256 in range(0, x):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "x",
    ),
    (
        """
@external
def repeat(n: uint256) -> uint256:
    for i: uint256 in range(0, n * 10):
        pass
    return n
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "n * 10",
    ),
    (
        """
@external
def bar():
    x:uint256 = 1
    for i: uint256 in range(0, x + 1):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "x + 1",
    ),
    (
        """
@external
def bar():
    for i: uint256 in range(2, 1):
        pass
    """,
        StructureException,
        "End must be greater than start",
        None,
        "1",
    ),
    (
        """
@external
def bar():
    x:uint256 = 1
    for i: uint256 in range(x, x):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "x",
    ),
    (
        """
@external
def foo():
    x: int128 = 5
    for i: int128 in range(x, x + 10):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "x",
    ),
    (
        """
@external
def repeat(n: uint256) -> uint256:
    for i: uint256 in range(n, 6):
        pass
    return x
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "n",
    ),
    (
        """
@external
def foo(x: int128):
    y: int128 = 7
    for i: int128 in range(x, x + y):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "x",
    ),
    (
        """
@external
def bar(x: uint256):
    for i: uint256 in range(3, x):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "x",
    ),
    (
        """
FOO: constant(int128) = 3
BAR: constant(uint256) = 7

@external
def foo():
    for i: uint256 in range(FOO, BAR):
        pass
    """,
        TypeMismatch,
        "Given reference has type int128, expected uint256",
        None,
        "FOO",
    ),
    (
        """
FOO: constant(int128) = -1

@external
def foo():
    for i: int128 in range(10, bound=FOO):
        pass
        """,
        StructureException,
        "Bound must be at least 1",
        None,
        "FOO",
    ),
    (
        """
@external
def foo():
    for i: DynArra[uint256, 3] in [1, 2, 3]:
        pass
    """,
        UnknownType,
        "No builtin or user-defined type named 'DynArra'.",
        "Did you mean 'DynArray'?",
        "DynArra",
    ),
    (
        # test for loop target broken into multiple lines
        """
@external
def foo():
    for i: \\
      \\
        \\
        \\
        \\
        \\
        uint9 in [1,2,3]:
        pass
    """,
        UnknownType,
        "No builtin or user-defined type named 'uint9'.",
        "Did you mean 'uint96', or maybe 'uint8'?",
        "uint9",
    ),
    (
        # test an even more deranged example
        """
@external
def foo():
    for i: \\
        \\
          DynArray[\\
        uint9, 3\\
        ] in [1,2,3]:
        pass
    """,
        UnknownType,
        "No builtin or user-defined type named 'uint9'.",
        "Did you mean 'uint96', or maybe 'uint8'?",
        "uint9",
    ),
    (
        """
@external
def foo():
    for i:decimal in range(1.1, 2.2):
        pass
    """,
        StructureException,
        "Value must be a literal integer, unless a bound is specified",
        None,
        "1.1",
    ),
    (
        """
@external
def foo():
    x:decimal = 1.1
    for i:decimal in range(x, x + 2.0, bound=10.1):
        pass
    """,
        StructureException,
        "Bound must be a literal integer",
        None,
        "10.1",
    ),
]

for_code_regex = re.compile(r"for .+ in (.*):", re.DOTALL)
fail_test_names = [
    (
        f"{i:02d}: {for_code_regex.search(code).group(1)}"  # type: ignore[union-attr]
        f" raises {type(err).__name__}"
    )
    for i, (code, err, msg, hint, src) in enumerate(fail_list)
]


@pytest.mark.parametrize(
    "bad_code,error_type,message,hint,source_code", fail_list, ids=fail_test_names
)
def test_range_fail(bad_code, error_type, message, hint, source_code):
    with pytest.raises(error_type) as exc_info:
        compiler.compile_code(bad_code)
    assert message == exc_info.value._message
    assert hint == exc_info.value.hint
    assert source_code == exc_info.value.args[1].get_original_node().node_source_code


valid_list = [
    """
@external
def foo():
    for i: uint256 in range(10):
        pass
    """,
    """
@external
def foo():
    for i: uint256 in range(10, 20):
        pass
    """,
    """
@external
def foo():
    x: int128 = 5
    for i: int128 in range(1, x, bound=4):
        pass
    """,
    """
@external
def foo():
    x: int128 = 5
    for i: int128 in range(x, bound=4):
        pass
    """,
    """
@external
def foo():
    x: int128 = 5
    for i: int128 in range(0, x, bound=4):
        pass
    """,
    """
interface Foo:
    def kick(): nonpayable
foos: Foo[3]
@external
def kick_foos():
    for foo: Foo in self.foos:
        extcall foo.kick()
    """,
]

valid_test_names = [
    f"{i} {for_code_regex.search(code).group(1)}"  # type: ignore[union-attr]
    for i, code in enumerate(valid_list)
]


@pytest.mark.parametrize("good_code", valid_list, ids=valid_test_names)
def test_range_success(good_code):
    assert compiler.compile_code(good_code) is not None



import pytest


def test_test_length(get_contract):
    test_length = """
y: Bytes[10]

@external
def foo(inp: Bytes[10]) -> uint256:
    x: Bytes[5] = slice(inp,1, 5)
    self.y = slice(inp, 2, 4)
    return len(inp) * 100 + len(x) * 10 + len(self.y)
    """

    c = get_contract(test_length)
    assert c.foo(b"badminton") == 954, c.foo(b"badminton")
    print("Passed length test")


@pytest.mark.parametrize("typ", ["DynArray[uint256, 50]", "Bytes[50]", "String[50]"])
def test_zero_length(get_contract, typ):
    code = f"""
@external
def boo() -> uint256:
    e: uint256 = len(empty({typ}))
    return e
    """

    c = get_contract(code)
    assert c.boo() == 0



import pytest

from vyper import compile_code
from vyper.ast.pre_parser import pre_parse, validate_version_pragma
from vyper.compiler.phases import CompilerData
from vyper.compiler.settings import OptimizationLevel, Settings
from vyper.exceptions import StructureException, VersionException

SRC_LINE = (1, 0)  # Dummy source line
COMPILER_VERSION = "0.1.1"
PRERELEASE_COMPILER_VERSION = "0.1.1b7"


@pytest.fixture
def mock_version(monkeypatch):
    def set_version(version):
        monkeypatch.setattr("vyper.__version__", version)

    return set_version


valid_versions = [
    "0.1.1",
    ">0.0.1",
    "^0.1.0",
    "<=1.0.0,>=0.1.0",
    # "0.1.0 - 1.0.0",
    "~=0.1.0",
]
invalid_versions = [
    "0.1.0",
    ">1.0.0",
    "^0.2.0",
    "<=0.0.1 >=1.1.1",
    "1.0.0 - 2.0.0",
    "~1.0.0",
    "0.2",
    "1",
    "1.x",
    "0.2.x",
    "0.2.0 || 0.1.3",
    "abc",
]


@pytest.mark.parametrize("file_version", valid_versions)
def test_valid_version_pragma(file_version, mock_version):
    mock_version(COMPILER_VERSION)
    validate_version_pragma(f"{file_version}", file_version, (SRC_LINE))


@pytest.mark.parametrize("file_version", invalid_versions)
def test_invalid_version_pragma(file_version, mock_version):
    mock_version(COMPILER_VERSION)
    with pytest.raises(VersionException):
        validate_version_pragma(f"{file_version}", file_version, (SRC_LINE))


prerelease_valid_versions = [
    "<0.1.1-beta.9",
    "<0.1.1b9",
    "0.1.1b7",
    ">0.1.1b2",
    "<0.1.1-rc.1",
    ">0.1.1a1",
    ">0.1.1-alpha.1",
    ">=0.1.1a9,<=0.1.1-rc.10",
    "<0.1.1b8",
    "<0.1.1rc1",
    "<0.2.0",
]
prerelease_invalid_versions = [
    ">0.1.1-beta.9",
    ">0.1.1b9",
    "0.1.1b8",
    "0.1.1rc2",
    "0.1.1-rc.9 - 0.1.1-rc.10",
    "<0.1.1b1",
    "<0.1.1a9",
]


@pytest.mark.parametrize("file_version", prerelease_valid_versions)
def test_prerelease_valid_version_pragma(file_version, mock_version):
    mock_version(PRERELEASE_COMPILER_VERSION)
    validate_version_pragma(file_version, file_version, (SRC_LINE))


@pytest.mark.parametrize("file_version", prerelease_invalid_versions)
def test_prerelease_invalid_version_pragma(file_version, mock_version):
    mock_version(PRERELEASE_COMPILER_VERSION)
    with pytest.raises(VersionException):
        validate_version_pragma(file_version, file_version, (SRC_LINE))


pragma_examples = [
    (
        """
    """,
        Settings(),
        Settings(optimize=OptimizationLevel.GAS),
    ),
    (
        """
    #pragma optimize codesize
    """,
        Settings(optimize=OptimizationLevel.CODESIZE),
        None,
    ),
    (
        """
    #pragma optimize none
    """,
        Settings(optimize=OptimizationLevel.NONE),
        None,
    ),
    (
        """
    #pragma optimize gas
    """,
        Settings(optimize=OptimizationLevel.GAS),
        None,
    ),
    (
        """
    #pragma version 0.3.10
    """,
        Settings(compiler_version="0.3.10"),
        Settings(optimize=OptimizationLevel.GAS),
    ),
    (
        """
    #pragma evm-version shanghai
    """,
        Settings(evm_version="shanghai"),
        Settings(evm_version="shanghai", optimize=OptimizationLevel.GAS),
    ),
    (
        """
    #pragma optimize codesize
    #pragma evm-version shanghai
    """,
        Settings(evm_version="shanghai", optimize=OptimizationLevel.CODESIZE),
        None,
    ),
    (
        """
    #pragma version 0.3.10
    #pragma evm-version shanghai
    """,
        Settings(evm_version="shanghai", compiler_version="0.3.10"),
        Settings(evm_version="shanghai", optimize=OptimizationLevel.GAS),
    ),
    (
        """
    #pragma version 0.3.10
    #pragma optimize gas
    """,
        Settings(compiler_version="0.3.10", optimize=OptimizationLevel.GAS),
        Settings(optimize=OptimizationLevel.GAS),
    ),
    (
        """
    #pragma version 0.3.10
    #pragma evm-version shanghai
    #pragma optimize gas
    """,
        Settings(compiler_version="0.3.10", optimize=OptimizationLevel.GAS, evm_version="shanghai"),
        Settings(optimize=OptimizationLevel.GAS, evm_version="shanghai"),
    ),
]


@pytest.mark.parametrize("code, pre_parse_settings, compiler_data_settings", pragma_examples)
def test_parse_pragmas(code, pre_parse_settings, compiler_data_settings, mock_version):
    mock_version("0.3.10")
    settings, _, _, _ = pre_parse(code)

    assert settings == pre_parse_settings

    compiler_data = CompilerData(code)

    # check what happens after CompilerData constructor
    if compiler_data_settings is None:
        # None is sentinel here meaning that nothing changed
        compiler_data_settings = pre_parse_settings

    # experimental_codegen is False by default
    compiler_data_settings.experimental_codegen = False

    assert compiler_data.settings == compiler_data_settings


invalid_pragmas = [
    # evm-versionnn
    """
# pragma evm-versionnn cancun
    """,
    # bad fork name
    """
# pragma evm-version cancunn
    """,
    # oppptimize
    """
# pragma oppptimize codesize
    """,
    # ggas
    """
# pragma optimize ggas
    """,
    # double specified
    """
# pragma optimize gas
# pragma optimize codesize
    """,
    # double specified
    """
# pragma evm-version cancun
# pragma evm-version shanghai
    """,
]


@pytest.mark.parametrize("code", invalid_pragmas)
def test_invalid_pragma(code):
    with pytest.raises(StructureException):
        pre_parse(code)


def test_version_exception_in_import(make_input_bundle):
    lib_version = "~=0.3.10"
    lib = f"""
#pragma version {lib_version}

@external
def foo():
    pass
    """

    code = """
import lib

uses: lib

@external
def bar():
    pass
    """
    input_bundle = make_input_bundle({"lib.vy": lib})

    with pytest.raises(VersionException) as excinfo:
        compile_code(code, input_bundle=input_bundle)
    annotation = excinfo.value.annotations[0]
    assert annotation.lineno == 2
    assert annotation.col_offset == 0
    assert annotation.full_source_code == lib



from vyper.codegen import core
from vyper.compiler.phases import CompilerData
from vyper.compiler.settings import OptimizationLevel, _is_debug_mode


def test_default_settings():
    source_code = ""
    compiler_data = CompilerData(source_code)
    _ = compiler_data.vyper_module  # force settings to be computed

    assert compiler_data.settings.optimize == OptimizationLevel.GAS


def test_default_opt_level():
    assert OptimizationLevel.default() == OptimizationLevel.GAS


def test_codegen_opt_level(optimize):
    assert core._opt_gas() == (optimize == OptimizationLevel.GAS)
    assert core._opt_none() == (optimize == OptimizationLevel.NONE)
    assert core._opt_codesize() == (optimize == OptimizationLevel.CODESIZE)


def test_debug_mode(pytestconfig):
    debug_mode = pytestconfig.getoption("enable_compiler_debug_mode")
    assert _is_debug_mode() == debug_mode



import functools
from typing import Any, Optional

from vyper import ast as vy_ast
from vyper.ast.validation import validate_call_args
from vyper.codegen.expr import Expr
from vyper.codegen.ir_node import IRnode
from vyper.exceptions import CompilerPanic, TypeMismatch, UnfoldableNode
from vyper.semantics.analysis.base import Modifiability
from vyper.semantics.analysis.utils import (
    check_modifiability,
    get_exact_type_from_node,
    validate_expected_type,
)
from vyper.semantics.types import TYPE_T, KwargSettings, VyperType
from vyper.semantics.types.utils import type_from_annotation


def process_arg(arg, expected_arg_type, context):
    # If the input value is a typestring, return the equivalent codegen type for IR generation
    if isinstance(expected_arg_type, TYPE_T):
        return expected_arg_type.typedef

    # if it is a word type, return a stack item.
    # TODO: remove this case, builtins should not require value expressions
    if expected_arg_type._is_prim_word:
        return Expr.parse_value_expr(arg, context)

    if isinstance(expected_arg_type, VyperType):
        return Expr(arg, context).ir_node

    raise CompilerPanic(f"Unexpected type: {expected_arg_type}")  # pragma: nocover


def process_kwarg(kwarg_node, kwarg_settings, expected_kwarg_type, context):
    if kwarg_settings.require_literal:
        return kwarg_node.get_folded_value().value

    return process_arg(kwarg_node, expected_kwarg_type, context)


def process_inputs(wrapped_fn):
    """
    Generate IR for input arguments on builtin functions.

    Applied as a wrapper on the `build_IR` method of
    classes in `vyper.functions.functions`.
    """

    @functools.wraps(wrapped_fn)
    def decorator_fn(self, node, context):
        subs = []
        for arg in node.args:
            arg_ir = process_arg(arg, arg._metadata["type"], context)
            # TODO annotate arg_ir with argname from self._inputs?
            subs.append(arg_ir)

        kwsubs = {}

        # note: must compile in source code order, left-to-right
        expected_kwarg_types = self.infer_kwarg_types(node)

        for k in node.keywords:
            kwarg_settings = self._kwargs[k.arg]
            expected_kwarg_type = expected_kwarg_types[k.arg]
            kwsubs[k.arg] = process_kwarg(k.value, kwarg_settings, expected_kwarg_type, context)

        # add kwargs which were not specified in the source
        for k, expected_arg in self._kwargs.items():
            if k not in kwsubs:
                kwsubs[k] = expected_arg.default

        for k, v in kwsubs.items():
            if isinstance(v, IRnode):
                v.annotation = k

        return wrapped_fn(self, node, subs, kwsubs, context)

    return decorator_fn


class BuiltinFunctionT(VyperType):
    typeclass = "builtin_function"

    _has_varargs = False
    _inputs: list[tuple[str, Any]] = []
    _kwargs: dict[str, KwargSettings] = {}
    _modifiability: Modifiability = Modifiability.MODIFIABLE
    _return_type: Optional[VyperType] = None
    _equality_attrs = ("_id",)
    _is_terminus = False

    @property
    def modifiability(self):
        return self._modifiability

    # helper function to deal with TYPE_Ts
    def _validate_single(self, arg: vy_ast.VyperNode, expected_type: VyperType) -> None:
        if TYPE_T.any().compare_type(expected_type):
            # try to parse the type - call type_from_annotation
            # for its side effects (will throw if is not a type)
            type_from_annotation(arg)
        else:
            validate_expected_type(arg, expected_type)

    def _validate_arg_types(self, node: vy_ast.Call) -> None:
        num_args = len(self._inputs)  # the number of args the signature indicates

        expect_num_args: Any = num_args
        if self._has_varargs:
            # note special meaning for -1 in validate_call_args API
            expect_num_args = (num_args, -1)

        validate_call_args(node, expect_num_args, list(self._kwargs.keys()))

        for arg, (_, expected) in zip(node.args, self._inputs):
            self._validate_single(arg, expected)

        for kwarg in node.keywords:
            kwarg_settings = self._kwargs[kwarg.arg]
            if kwarg_settings.require_literal and not check_modifiability(
                kwarg.value, Modifiability.CONSTANT
            ):
                raise TypeMismatch("Value must be literal", kwarg.value)
            self._validate_single(kwarg.value, kwarg_settings.typ)

        # typecheck varargs. we don't have type info from the signature,
        # so ensure that the types of the args can be inferred exactly.
        varargs = node.args[num_args:]
        if len(varargs) > 0:
            assert self._has_varargs  # double check validate_call_args
        for arg in varargs:
            # call get_exact_type_from_node for its side effects -
            # ensures the type can be inferred exactly.
            get_exact_type_from_node(arg)

    def check_modifiability_for_call(self, node: vy_ast.Call, modifiability: Modifiability) -> bool:
        return self._modifiability <= modifiability

    def fetch_call_return(self, node: vy_ast.Call) -> Optional[VyperType]:
        self._validate_arg_types(node)

        return self._return_type

    def infer_arg_types(self, node: vy_ast.Call, expected_return_typ=None) -> list[VyperType]:
        self._validate_arg_types(node)
        ret = [expected for (_, expected) in self._inputs]

        # handle varargs.
        n_known_args = len(self._inputs)
        varargs = node.args[n_known_args:]
        if len(varargs) > 0:
            assert self._has_varargs
        ret.extend(get_exact_type_from_node(arg) for arg in varargs)
        return ret

    def infer_kwarg_types(self, node: vy_ast.Call) -> dict[str, VyperType]:
        return {i.arg: self._kwargs[i.arg].typ for i in node.keywords}

    def __repr__(self):
        return f"(builtin) {self._id}"

    def _try_fold(self, node):
        raise UnfoldableNode(f"not foldable: {self}", node)



import pytest

from tests.utils import ZERO_ADDRESS

# ERC1155 ownable, opensea compatible tests
# @author Dr. Pixel (github: @Doc-Pixel)

# constants - contract deployment
CONTRACT_NAME = "TEST 1155"
CONTRACT_SYMBOL = "T1155"

CONTRACT_URI = "https://mydomain.io/NFTdata/ID"
NEW_CONTRACT_URI = "https://mynewdomain.io/NFTdata/ID"
CONTRACT_METADATA_URI = "https://mydomain.io/NFTdata/collectionMetaData.json"
NEW_CONTRACT_METADATA_URI = "https://mydomain.io/NFTdata/newCollectionMetaData.json"

CONTRACT_DYNURI = "https://mydomain.io/NFTdata/"


ERC165_INTERFACE_ID = "0x01ffc9a7"
ERC1155_INTERFACE_ID = "0xd9b67a26"
ERC1155_INTERFACE_ID_METADATA = "0x0e89341c"
DUMMY_BYTES32_DATA = "0x0101010101010101010101010101010101010101010101010101010101010101"

# minting test lists
mintBatch = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
mintBatch2 = [11, 12, 13, 14, 15, 16, 17, 19, 19, 20]
minBatchSetOf10 = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
mintConflictBatch = [1, 2, 3]


@pytest.fixture(scope="module")
def erc1155(get_contract, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    with open("examples/tokens/ERC1155ownable.vy") as f:
        code = f.read()
    c = get_contract(code, *[CONTRACT_NAME, CONTRACT_SYMBOL, CONTRACT_URI, CONTRACT_METADATA_URI])
    assert c.owner() == owner
    c.mintBatch(a1, mintBatch, minBatchSetOf10, sender=owner)
    c.mintBatch(a3, mintBatch2, minBatchSetOf10, sender=owner)

    assert c.balanceOf(a1, 1) == 1
    assert c.balanceOf(a1, 2) == 1
    assert c.balanceOf(a1, 3) == 1
    with tx_failed():
        c.mintBatch(ZERO_ADDRESS, mintBatch, minBatchSetOf10, sender=owner)
    with tx_failed():
        c.mintBatch(a1, [1, 2, 3], [1, 1], sender=owner)

    c.mint(a1, 21, 1, sender=owner)
    c.mint(a1, 22, 1, sender=owner)
    c.mint(a1, 23, 1, sender=owner)
    c.mint(a1, 24, 1, sender=owner)

    with tx_failed():
        c.mint(a1, 24, 1, sender=a3)
    with tx_failed():
        c.mint(ZERO_ADDRESS, 24, 1, sender=owner)

    assert c.balanceOf(a1, 21) == 1
    assert c.balanceOf(a1, 22) == 1
    assert c.balanceOf(a1, 23) == 1
    assert c.balanceOf(a1, 24) == 1

    return c


# tests


def test_initial_state(erc1155):
    # Check if the constructor of the contract is set up properly
    # and the contract is deployed with the desired variables

    # variables set correctly?
    assert erc1155.name() == CONTRACT_NAME
    assert erc1155.symbol() == CONTRACT_SYMBOL
    assert erc1155.uri(0) == CONTRACT_URI

    # interfaces set up correctly?
    assert erc1155.supportsInterface(ERC165_INTERFACE_ID)
    assert erc1155.supportsInterface(ERC1155_INTERFACE_ID)
    assert erc1155.supportsInterface(ERC1155_INTERFACE_ID_METADATA)


def test_pause(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    # check the pause status, pause, check, unpause, check, with owner and non-owner accounts
    # this test will check all the function that should not work when paused.
    assert not erc1155.paused()

    # try to pause the contract from a non owner account
    with tx_failed():
        erc1155.pause(sender=a1)

    # now pause the contract and check status
    erc1155.pause(sender=owner)
    assert erc1155.paused()

    # try pausing a paused contract
    with tx_failed():
        erc1155.pause()

    # try functions that should not work when paused
    with tx_failed():
        erc1155.setURI(NEW_CONTRACT_URI)

    # test burn and burnbatch
    with tx_failed():
        erc1155.burn(21, 1)
    with tx_failed():
        erc1155.burnBatch([21, 22], [1, 1])

    # check mint and mintbatch
    with tx_failed():
        erc1155.mint(a1, 21, 1, sender=owner)
    with tx_failed():
        erc1155.mintBatch(a1, mintBatch, minBatchSetOf10, sender=owner)

    # check safetransferfrom and safebatchtransferfrom
    with tx_failed():
        erc1155.safeTransferFrom(a1, a2, 21, 1, DUMMY_BYTES32_DATA, sender=a1)
    with tx_failed():
        erc1155.safeBatchTransferFrom(
            a1, a2, [21, 22, 23], [1, 1, 1], DUMMY_BYTES32_DATA, sender=a1
        )

    # check ownership functions
    with tx_failed():
        erc1155.transferOwnership(a1)
    with tx_failed():
        erc1155.renounceOwnership()

    # check approval functions
    with tx_failed():
        erc1155.setApprovalForAll(owner, a5, True)

    # try and unpause as non-owner
    with tx_failed():
        erc1155.unpause(sender=a1)

    erc1155.unpause(sender=owner)
    assert not erc1155.paused()

    # try un pausing an unpaused contract
    with tx_failed():
        erc1155.unpause()


def test_contractURI(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    # change contract URI and restore.
    assert erc1155.contractURI() == CONTRACT_METADATA_URI
    with tx_failed():
        erc1155.setContractURI(NEW_CONTRACT_METADATA_URI, sender=a1)
    erc1155.setContractURI(NEW_CONTRACT_METADATA_URI, sender=owner)
    assert erc1155.contractURI() == NEW_CONTRACT_METADATA_URI
    assert erc1155.contractURI() != CONTRACT_METADATA_URI
    erc1155.setContractURI(CONTRACT_METADATA_URI, sender=owner)
    assert erc1155.contractURI() != NEW_CONTRACT_METADATA_URI
    assert erc1155.contractURI() == CONTRACT_METADATA_URI

    with tx_failed():
        erc1155.setContractURI(CONTRACT_METADATA_URI)


def test_URI(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    # change contract URI and restore.
    assert erc1155.uri(0) == CONTRACT_URI
    erc1155.setURI(NEW_CONTRACT_URI, sender=owner)
    assert erc1155.uri(0) == NEW_CONTRACT_URI
    assert erc1155.uri(0) != CONTRACT_URI
    erc1155.setURI(CONTRACT_URI, sender=owner)
    assert erc1155.uri(0) != NEW_CONTRACT_URI
    assert erc1155.uri(0) == CONTRACT_URI

    with tx_failed():
        erc1155.setURI(CONTRACT_URI)

    # set contract to dynamic URI
    erc1155.toggleDynUri(True, sender=owner)
    erc1155.setURI(CONTRACT_DYNURI, sender=owner)
    assert erc1155.uri(0) == CONTRACT_DYNURI + str(0) + ".json"


def test_safeTransferFrom_balanceOf_single(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    assert erc1155.balanceOf(a1, 24) == 1
    # transfer by non-owner
    with tx_failed():
        erc1155.safeTransferFrom(a1, a2, 24, 1, DUMMY_BYTES32_DATA, sender=a2)

    # transfer to zero address
    with tx_failed():
        erc1155.safeTransferFrom(a1, ZERO_ADDRESS, 24, 1, DUMMY_BYTES32_DATA, sender=a1)

    # transfer to self
    with tx_failed():
        erc1155.safeTransferFrom(a1, a1, 24, 1, DUMMY_BYTES32_DATA, sender=a1)

    # transfer more than owned
    with tx_failed():
        erc1155.safeTransferFrom(a1, a2, 24, 500, DUMMY_BYTES32_DATA, sender=a1)

    # transfer item not owned / not existing
    with tx_failed():
        erc1155.safeTransferFrom(a1, a2, 500, 1, DUMMY_BYTES32_DATA, sender=a1)

    erc1155.safeTransferFrom(a1, a2, 21, 1, DUMMY_BYTES32_DATA, sender=a1)

    assert erc1155.balanceOf(a2, 21) == 1

    # try to transfer item again
    with tx_failed():
        erc1155.safeTransferFrom(a1, a2, 21, 1, DUMMY_BYTES32_DATA, sender=a1)
    assert erc1155.balanceOf(a1, 21) == 0


# TODO: mint 20 NFTs [1:20] and check the balance for each
def test_mintBatch_balanceOf(erc1155, env, tx_failed):  # test_mint_batch
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    # Use the mint three fixture to mint the tokens.
    # this test checks the balances of this test
    for i in range(1, 10):
        assert erc1155.balanceOf(a1, i) == 1


def test_safeBatchTransferFrom_balanceOf_batch(erc1155, env, tx_failed):  # test_mint_batch
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]

    # check a1 balances for NFTs 21-24
    assert erc1155.balanceOf(a1, 21) == 1
    assert erc1155.balanceOf(a1, 22) == 1
    assert erc1155.balanceOf(a1, 23) == 1
    assert erc1155.balanceOf(a1, 23) == 1

    # try to transfer item from non-item owner account
    with tx_failed():
        erc1155.safeBatchTransferFrom(
            a1, a2, [21, 22, 23], [1, 1, 1], DUMMY_BYTES32_DATA, sender=a2
        )

    # try to transfer item to zero address
    with tx_failed():
        erc1155.safeBatchTransferFrom(
            a1, ZERO_ADDRESS, [21, 22, 23], [1, 1, 1], DUMMY_BYTES32_DATA, sender=a1
        )

    # try to transfer item to self
    with tx_failed():
        erc1155.safeBatchTransferFrom(
            a1, a1, [21, 22, 23], [1, 1, 1], DUMMY_BYTES32_DATA, sender=a1
        )

    # try to transfer more items than we own
    with tx_failed():
        erc1155.safeBatchTransferFrom(
            a1, a2, [21, 22, 23], [1, 125, 1], DUMMY_BYTES32_DATA, sender=a1
        )

    # mismatched item and amounts
    with tx_failed():
        erc1155.safeBatchTransferFrom(a1, a2, [21, 22, 23], [1, 1], DUMMY_BYTES32_DATA, sender=a1)

    # try to transfer nonexisting item
    with tx_failed():
        erc1155.safeBatchTransferFrom(
            a1, a2, [21, 22, 500], [1, 1, 1], DUMMY_BYTES32_DATA, sender=a1
        )
    erc1155.safeBatchTransferFrom(a1, a2, [21, 22, 23], [1, 1, 1], DUMMY_BYTES32_DATA, sender=a1)

    # try to transfer again, our balances are zero now, should fail
    with tx_failed():
        erc1155.safeBatchTransferFrom(
            a1, a2, [21, 22, 23], [1, 1, 1], DUMMY_BYTES32_DATA, sender=a1
        )
    with tx_failed():
        erc1155.balanceOfBatch([a2, a2, a2], [21, 22], sender=owner)

    assert erc1155.balanceOfBatch([a2, a2, a2], [21, 22, 23]) == [1, 1, 1]
    assert erc1155.balanceOf(a1, 21) == 0


def test_mint_one_burn_one(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]

    # check the balance from an owner and non-owner account
    erc1155.mint(owner, 25, 1, sender=owner)

    assert erc1155.balanceOf(owner, 25) == 1
    assert erc1155.balanceOf(owner, 25) == 1

    # try and burn an item we don't control
    with tx_failed():
        erc1155.burn(25, 1, sender=a3)

    # burn an item that contains something we don't own
    with tx_failed():
        erc1155.burn(595, 1, sender=a1)

    # burn ah item passing a higher amount than we own
    with tx_failed():
        erc1155.burn(25, 500, sender=a1)

    erc1155.burn(25, 1, sender=owner)

    assert erc1155.balanceOf(owner, 25) == 0


def test_mint_batch_burn_batch(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    # mint NFTs 11-20

    # check the balance
    assert erc1155.balanceOfBatch([a3, a3, a3], [11, 12, 13]) == [1, 1, 1]

    # try and burn a batch we don't control
    with tx_failed():
        erc1155.burnBatch([11, 12], [1, 1])

    # ids and amounts array length not matching
    with tx_failed():
        erc1155.burnBatch([1, 2, 3], [1, 1], sender=a1)

    # burn a batch that contains something we don't own
    with tx_failed():
        erc1155.burnBatch([2, 3, 595], [1, 1, 1], sender=a1)

    # burn a batch passing a higher amount than we own
    with tx_failed():
        erc1155.burnBatch([1, 2, 3], [1, 500, 1], sender=a1)

    # burn existing
    erc1155.burnBatch([11, 12], [1, 1], sender=a3)

    assert erc1155.balanceOfBatch([a3, a3, a3], [11, 12, 13]) == [0, 0, 1]

    # burn again, should revert
    with tx_failed():
        erc1155.burnBatch([11, 12], [1, 1], sender=a3)

    assert erc1155.balanceOfBatch([a3, a3, a3], [1, 2, 3]) == [0, 0, 0]


def test_approval_functions(erc1155, env, tx_failed):  # test_mint_batch
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    # self-approval by the owner
    with tx_failed():
        erc1155.setApprovalForAll(a5, a5, True, sender=a5)

    # let's approve and operator for somebody else's account
    with tx_failed():
        erc1155.setApprovalForAll(owner, a5, True, sender=a3)

    # set approval correctly
    erc1155.setApprovalForAll(owner, a5, True)

    # check approval
    erc1155.isApprovedForAll(owner, a5)

    # remove approval
    erc1155.setApprovalForAll(owner, a5, False)


def test_max_batch_size_violation(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    TOTAL_BAD_BATCH = 200
    ids = []
    amounts = []
    for i in range(1, TOTAL_BAD_BATCH):
        ids.append(i)
        amounts.append(1)

    with tx_failed():
        erc1155.mintBatch(a1, ids, amounts, sender=owner)


# Transferring back and forth


def test_ownership_functions(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    print(owner, a1, a2)
    print("___owner___", erc1155.owner())
    # change owner from account 0 to account 1 and back

    assert erc1155.owner() == owner
    with tx_failed():
        erc1155.transferOwnership(a1, sender=a2)

    # try to transfer ownership to current owner
    with tx_failed():
        erc1155.transferOwnership(owner)
    # try to transfer ownership to ZERO ADDRESS
    with tx_failed():
        erc1155.transferOwnership(ZERO_ADDRESS)

    # Transfer ownership to account 1
    erc1155.transferOwnership(a1, sender=owner)
    # assert erc1155.owner() == a1
    assert erc1155.owner() == a1


def test_renounce_ownership(erc1155, env, tx_failed):
    owner, a1, a2, a3, a4, a5 = env.accounts[0:6]
    assert erc1155.owner() == owner
    # try to transfer ownership from non-owner account
    with tx_failed():
        erc1155.renounceOwnership(sender=a2)

    erc1155.renounceOwnership(sender=owner)

    # assert erc1155.owner() == ZERO_ADDRESS



import pytest

from vyper import compiler
from vyper.exceptions import TypeMismatch

fail_list = [
    (
        """
@external
def foo() -> int128[2]:
    return [3,chain.id]
    """,
        TypeMismatch,
    ),
    """
@external
def foo() -> decimal:
    x: int128 = as_wei_value(5, "finney")
    y: int128 = chain.id + 50
    return x / y
    """,
    """
@external
def foo():
    x: int128 = 7
    y: int128 = min(x, chain.id)
    """,
    """
a: HashMap[uint256, int128]

@external
def add_record():
    self.a[chain.id] = chain.id + 20
    """,
    """
a: HashMap[int128, uint256]

@external
def add_record():
    self.a[chain.id] = chain.id + 20
    """,
    (
        """
@external
def foo(inp: Bytes[10]) -> Bytes[3]:
    return slice(inp, chain.id, -3)
    """,
        TypeMismatch,
    ),
]


@pytest.mark.parametrize("bad_code", fail_list)
def test_chain_fail(bad_code):
    if isinstance(bad_code, tuple):
        with pytest.raises(bad_code[1]):
            compiler.compile_code(bad_code[0])
    else:
        with pytest.raises(TypeMismatch):
            compiler.compile_code(bad_code)


valid_list = [
    """
@external
@view
def get_chain_id() -> uint256:
    return chain.id
    """,
    """
@external
@view
def check_chain_id(c: uint256) -> bool:
    return chain.id == c
    """,
]


@pytest.mark.parametrize("good_code", valid_list)
def test_chain_success(good_code):
    assert compiler.compile_code(good_code) is not None


def test_chainid_operation(get_contract, env):
    code = """
@external
@view
def get_chain_id() -> uint256:
    return chain.id
    """
    c = get_contract(code)
    assert c.get_chain_id() == env.DEFAULT_CHAIN_ID
